from imnn.imnn import IMNN, SimulatorIMNN, AggregatedSimulatorIMNN, \
    GradientIMNN, AggregatedGradientIMNN, DatasetGradientIMNN, \
    NumericalGradientIMNN, AggregatedNumericalGradientIMNN, \
    DatasetNumericalGradientIMNN
from imnn.utils import TFRecords

__author__ = "Tom Charnock"
__version__ = "0.3.2"

__all__ = [
    "IMNN",
    "SimulatorIMNN",
    "AggregatedSimulatorIMNN",
    "GradientIMNN",
    "AggregatedGradientIMNN",
    "DatasetGradientIMNN",
    "NumericalGradientIMNN",
    "AggregatedNumericalGradientIMNN",
    "DatasetNumericalGradientIMNN",
    "TFRecords"
]



from imnn.utils.container import container
from imnn.utils.jac import jacrev, value_and_jacrev, jacfwd, \
    value_and_jacfwd
from imnn.utils.utils import add_nested_pytrees
from imnn.utils.progress_bar import progress_bar
from imnn.utils.tfrecords import TFRecords

__author__ = "Tom Charnock"
__version__ = "0.3.2"

__all__ = [
    "container",
    "jacrev",
    "value_and_jacrev",
    "jacfwd",
    "value_and_jacfwd",
    "progress_bar",
    "add_nested_pytrees",
    "TFRecords"
]



import pytest
import jax.numpy as np
from test.simulator_defaults import simulatorTests
from imnn.imnn import SimulatorIMNN

test = simulatorTests(imnn=SimulatorIMNN, filename="simulator")


def fail_fn(a, b, c):
    pass


# Test that all initialisation parameters correctly raise errors
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize(
    "input_variable",
    [None, list(), 1., 1, np.zeros((1,)), test.rng, tuple(), (0, 0),
     (test.model[0], 0), test.bad_model, test.state, fail_fn])
@pytest.mark.parametrize("variable", test.kwargs.keys())
def test_initialisation_parameters_(variable, kwargs, input_variable, state):
    test.initialise_parameters(
        variable, kwargs, input_variable, state=state, validate=False)


# Test that all parameters passed to fit correctly raise errors
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize(
    "input_variable", [None, list(), 1., 1, np.zeros((1,)), test.rng])
@pytest.mark.parametrize("variable", test.fit_kwargs.keys())
def test_fit_parameters_(variable, kwargs, input_variable, state):
    test.fit_parameters(
        variable, kwargs, test.fit_kwargs, input_variable, state=state,
        validate=False)


# Test that fitting correctly fails and get_estimate won't run and that plot
# can be made
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("none_first", [True, False])
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
def test_combined_running_test_(kwargs, state, validate, fit, none_first):
    test.combined_running_test(
        [test.single_target_data, test.batch_target_data], kwargs,
        test.fit_kwargs, state=state, validate=validate, fit=fit,
        none_first=none_first)



import numpy as np
from imnn.imnn import AggregatedNumericalGradientIMNN
from imnn.utils.utils import _check_input, _check_type


class DatasetNumericalGradientIMNN(AggregatedNumericalGradientIMNN):
    """Information maximising neural network fit using numerical derivatives

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    simulations :math:`{\\bf d}^i` originally generated at fiducial model
    parameter :math:`{\\bf\\theta}^\\rm{fid}`, and a set of
    :math:`i\\in[1, n_d]` simulations,
    :math:`\\{{\\bf d}_{\\alpha^-}^i, {\\bf d}_{\\alpha^+}^i\\}`, generated
    with the same seed at each :math:`i` generated at
    :math:`{\\bf\\theta}^\\rm{fid}` apart from at parameter label
    :math:`\\alpha` with values

    .. math::
        \\theta_{\\alpha^-} = \\theta_\\alpha^\\rm{fid}-\\delta\\theta_\\alpha

    and

    .. math::
        \\theta_{\\alpha^+} = \\theta_\\alpha^\\rm{fid}+\\delta\\theta_\\alpha

    where :math:`\\delta\\theta_\\alpha` is a :math:`n_{params}` length vector
    with the :math:`\\alpha` element having a value which perturbs the
    parameter :math:`\\theta^{\\rm fid}_\\alpha`. This means there are
    :math:`2\\times n_{params}\\times n_d` simulations used to calculate the
    numerical derivatives (this is extremely cheap compared to other machine
    learning methods). All these simulations are passed through a network
    :math:`f_{{\\bf w}}({\\bf d})` with network parameters :math:`{\\bf w}` to
    obtain network outputs :math:`{\\bf x}^i` and
    :math:`\\{{\\bf x}_{\\alpha^-}^i,{\\bf x}_{\\alpha^+}^i\\}`. These
    perturbed values are combined to obtain

    .. math::
        \\frac{\\partial{{\\bf x}^i}}{\\partial\\theta_\\alpha} =
        \\frac{{\\bf x}_{\\alpha^+}^i - {\\bf x}_{\\alpha^-}^i}
        {\\delta\\theta_\\alpha}

    With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    To enable the use of large data (or networks) the whole procedure is
    aggregated. This means that the passing of the simulations through the
    network is farmed out to the desired XLA devices, and recollected,
    ``n_per_device`` inputs at a time. These are then used to calculate the
    automatic gradient of the loss function with respect to the calculated
    summaries and derivatives, :math:`\\partial\\Lambda/\\partial{\\bf x}^i`
    (which is a fairly small computation as long as ``n_summaries`` and ``n_s``
    {and ``n_d``} are not huge). Once this is calculated, the simulations are
    passed through the network AGAIN this time calculating the Jacobian of the
    network output with respect to the network parameters
    :math:`\\partial{\\bf x}^i/\\partial{\\bf w}` which is then combined via
    the chain rule to get

    .. math::
        \\frac{\\partial\\Lambda}{\\partial{\\bf w}} =
        \\frac{\\partial\\Lambda}{\\partial{\\bf x}^i}
        \\frac{\\partial{\\bf x}^i}{\\partial{\\bf w}}

    This can then be passed to the optimiser.

    In DatasetNumericalGradientIMNN the input datasets should be lists of
    ``n_devices`` ``tf.data.Datasets``. Please note, due to the many various
    ways of constructing datasets to load data, there is no checking and any
    improperly made dataset will either fail (best result) or provide the wrong
    result (worst case scenario!). For this reason it is advised to use
    :func:`~AggregatedNumericalGradientIMNN` if data will fit into CPU memory
    at least. If not, the next safest way is to construct a set of TFRecords
    and construct the dataset from that.

    Examples
    --------
    Here are various ways to construct the datasets for passing to the
    :func:~`DatasetNumericalGradientIMNN`. Note that these are not the only
    ways, but they should give something to follow to generate your own
    datasets. First we'll generate some data (just random noise with zero mean
    and unit variance) and perturb the mean and the variance of this noise to
    calculate numerical derivatives with respect to the model parameters.
    We'll generate 1000 simulations at the fiducial and 100 for each parameter
    varied above and below the fiducial. We'll save each of these simulations
    into its own individual file (named by seed value).

    .. code-block:: python

        import glob
        import jax
        import jax.numpy as np
        import tensorflow as tf
        from imnn import TFRecords
        from functools import partial

        n_s = 1000
        n_d = 100
        n_params = 2
        input_shape = (10,)

        def simulator(key, θ):
            return θ[0] + (jax.random.normal(key, shape=input_shape)
                * np.sqrt(θ[1]))

        θ_fid = np.array([0., 1.])
        δθ = np.array([0.1, 0.1])
        θ_der = (θ_fid
            + np.einsum(
                "i,jk->ijk",
                np.array([-1., 1.]),
                np.diag(δθ)
            / 2.)).reshape((-1, 2))

        rng = jax.random.PRNGKey(0)
        rng, data_key = jax.random.split(rng)
        data_keys = np.array(jax.random.split(rng, num=2 * n_s))

        fiducial = jax.vmap(simulator)(
            data_keys[:n_s],
            np.repeat(np.expand_dims(θ_fid, 0), n_s, axis=0))

        validation_fiducial = jax.vmap(simulator)(
            data_keys[n_s:],
            np.repeat(np.expand_dims(θ_fid, 0), n_s, axis=0))

        numerical_derivative = jax.vmap(simulator)(
            np.repeat(data_keys[:n_d], θ_der.shape[0], axis=0),
            np.tile(θ_der, (n_d, 1))).reshape(
                (n_d, 2, n_params) + input_shape)

        validation_numerical_derivative = jax.vmap(simulator)(
            np.repeat(data_keys[n_s:n_d + n_s], θ_der.shape[0], axis=0),
            np.tile(θ_der, (n_d, 1))).reshape(
                (n_d, 2, n_params) + input_shape)

        for i, (simulation, validation_simulation) in enumerate(
                zip(fiducial, validation_fiducial)):
            np.save(f"tmp/fiducial_{i:04d}.npy", simulation)
            np.save(f"tmp/validation_fiducial_{i:04d}.npy",
                    validation_simulation)

        for i, (simulation, validation_simulation) in enumerate(
                zip(numerical_derivative, validation_numerical_derivative)):
            np.save(f"tmp/numerical_derivative_{i:04d}.npy", simulation)
            np.save(f"tmp/validation_numerical_derivative_{i:04d}.npy",
                    validation_simulation

    Now we'll define how many devices to farm out our calculations to. We need
    to know this because we want to make a separate dataset for each device.
    We'll also set the number of simulations which can be processed at once on
    each device, this should be as high as possible without running out of
    memory on any individual device for quickest fitting

    .. code-block:: python

        devices = jax.devices("gpu")
        n_devices = len(devices)
        n_per_device = 100

    Using this we can define the shapes for the data to be reshaped into for
    proper construction of the datasets to be used when fitting the IMNN.

    .. code-block:: python

        fiducial_shape = (
            n_devices,
            n_s // (n_devices * n_per_device),
            n_per_device) + input_shape

        derivative_shape = (
            n_devices,
            2 * n_params * n_d // (n_devices * n_per_device),
            n_per_device) + input_shape

    The simplest way to construct a dataset is simply using the numpy arrays
    from memory (note if you're going to do this you should really just use
    ``AggregatedNumericalGradientIMNN``, its more or less the same!), i.e.

    .. code-block:: python

        fiducial = [
            tf.data.Dataset.from_tensor_slices(
                fid).repeat().as_numpy_iterator()
            for fid in fiducial.reshape(fiducial_shape)]

        numerical_derivative = [
            tf.data.Dataset.from_tensor_slices(
                der).repeat().as_numpy_iterator()
            for der in numerical_derivative.reshape(derivative_shape)]

        validation_fiducial = [
            tf.data.Dataset.from_tensor_slices(
                fid).repeat().as_numpy_iterator()
            for fid in validation_fiducial.reshape(fiducial_shape)]

        validation_numerical_derivative = [
            tf.data.Dataset.from_tensor_slices(
                der).repeat().as_numpy_iterator()
            for der in validation_numerical_derivative.reshape(
                derivative_shape)]

    However, if the data is too large to fit in memory then we can use the
    npy files that we saved by loading them via a generator

    .. code-block:: python

        def generator(directory, filename, total):
            i = 0
            while i < total:
                yield np.load(f"{directory}/{filename}_{i:04d}.npy")
                i += 1

    We can then build the datasets like:

    .. code-block:: python

        fiducial = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "fiducial",
                    n_s),
                tf.float32
                ).take(n_s // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        numerical_derivative = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "numerical_derivative",
                    n_d),
                tf.float32
                ).flat_map(
                    lambda x: tf.data.Dataset.from_tensor_slices(x)
                ).flat_map(
                    lambda x: tf.data.Dataset.from_tensor_slices(x)
                ).take(2 * n_params * n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        validation_fiducial = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "validation_fiducial",
                    n_s),
                tf.float32
                ).take(n_s // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        validation_numerical_derivative = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "validation_numerical_derivative",
                    n_d),
                tf.float32
                ).flat_map(
                    lambda x: tf.data.Dataset.from_tensor_slices(x)
                ).flat_map(
                    lambda x: tf.data.Dataset.from_tensor_slices(x)
                ).take(2 * n_params * n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

    The datasets must be built exactly like this, with the taking and batching
    and repeating. Importantly both the ``flat_map`` over the datasets for the
    numerical derivatives are needed to unwrap the perturbation direction and
    the parameter direction in each numpy file. To prefetch and cache the
    loaded files we can add the extra steps in the datasets, e.g.

    .. code-block:: python

        fiducial = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "fiducial",
                    n_s),
                tf.float32
                ).take(n_s // n_devices
                ).batch(n_per_device
                ).cache(
                ).prefetch(tf.data.AUTOTUNE
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

    This loading will be quite slow because the files need to be opened each
    time, but we can build TFRecords which are quicker to load. There is a
    writer able to do the correct format. The TFRecords should be a couple
    hundred Mb for best flow-through, so we can keep filling the record until
    this size is reached.

    .. code-block:: python

        record_size = 200 #Mb
        writer = TFRecords.TFRecords(record_size=record_size)

    We need a function which grabs single simulations from an array (or file)
    to add to the record

    .. code-block:: python

        def get_fiducial(seed, directory=None, filename=None):
            return np.load(f"{directory}/{filename}_{seed:04d}.npy")

        def get_derivative(seed, der, params, directory=None, filename=None):
            return np.load(
                f"{directory}/{filename}_{seed:04d}.npy")[der, params]

        writer.write_record(
            n_sims=n_s,
            get_simulation=lambda seed: get_fiducial(
                seed, directory="tmp", filename="fiducial"),
            fiducial=True,
            directory="tmp",
            filename="fiducial")

        writer.write_record(
            n_sims=n_d,
            get_simulation=lambda seed, der, param: get_derivative(
                seed, der, param, directory="tmp",
                filename="numerical_derivative"),
            fiducial=False,
            n_params=n_params,
            directory="tmp",
            filename="numerical_derivative")

        writer.write_record(
            n_sims=n_s,
                get_simulation=lambda seed: get_fiducial(
                seed, directory="tmp",
                filename="validation_fiducial"),
            fiducial=True,
            directory="tmp",
            filename="validation_fiducial")

        writer.write_record(
            n_sims=n_d,
            get_simulation=lambda seed, der, param: get_derivative(
                seed, der, param, directory="tmp",
                filename="validation_numerical_derivative"),
            fiducial=False,
            n_params=n_params,
            directory="tmp",
            filename="validation_numerical_derivative")

    We can then read these to a dataset using the parser from the TFRecords
    class mapping the format of the data to a 32-bit float

    .. code-block:: python

        fiducial = [
            tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/fiducial_*.tfrecords")),
                    num_parallel_reads=1
                ).map(writer.parser
                ).skip(i * n_s // n_devices
                ).take(n_s // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for i in range(n_devices)]

        numerical_derivative = [
            tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/numerical_derivative_*.tfrecords")),
                    num_parallel_reads=1
                ).map(writer.parser
                ).skip(i * 2 * n_params * n_d // n_devices
                ).take(2 * n_params * n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for i in range(n_devices)]

        validation_fiducial = [
            tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/validation_fiducial_*.tfrecords")),
                    num_parallel_reads=1
                ).map(writer.parser
                ).skip(i * n_s // n_devices
                ).take(n_s // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        validation_numerical_derivative = [
            tf.data.TFRecordDataset(
                    sorted(glob.glob(
                        "tmp/validation_numerical_derivative_*.tfrecords")),
                    num_parallel_reads=1
                ).map(writer.parser
                ).skip(i * 2 * n_params * n_d // n_devices
                ).take(2 * n_params * n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

    Parameters
    ----------
    δθ : float(n_params,)
        Size of perturbation to model parameters for the numerical derivative
    fiducial : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for fitting). These
        are served ``n_per_device`` at a time as a numpy iterator from a
        TensorFlow dataset.
    derivative : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network outputs
        with respect to model parameters (for fitting).  These are served
        ``n_per_device`` at a time as a numpy iterator from a TensorFlow
        dataset.
    validation_fiducial : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for validation).
        These are served ``n_per_device`` at a time as a numpy iterator from a
        TensorFlow dataset.
    validation_derivative : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network outputs
        with respect to model parameters (for validation).  These are served
        ``n_per_device`` at a time as a numpy iterator from a TensorFlow
        dataset.
    fiducial_iterations : int
        The number of iterations over the fiducial dataset
    derivative_iterations : int
        The number of iterations over the derivative dataset
    derivative_output_shape : tuple
        The shape of the output of the derivatives from the network
    fiducial_batch_shape : tuple
        The shape of each batch of fiducial simulations (without input or
        summary shape)
    derivative_batch_shape : tuple
        The shape of each batch of derivative simulations (without input or
        summary shape)
    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, fiducial, derivative, δθ,
                 host, devices, n_per_device, validation_fiducial=None,
                 validation_derivative=None):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also fills the
        simulation attributes (and validation if available).

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        fiducial : list of tf.data.Dataset()
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs (for
            fitting). These are served ``n_per_device`` at at time as a numpy
            iterator from a TensorFlow dataset.
        derivative : list of tf.data.Dataset()
            The simulations generated at parameter values perturbed from the
            fiducial used to calculate the numerical derivative of network
            outputs with respect to model parameters (for fitting).  These are
            served ``n_per_device`` at at time as a numpy iterator from a
            TensorFlow dataset.
        δθ : float(n_params,)
            Size of perturbation to model parameters for the numerical
            derivative
        host: jax.device
            The main device where the Fisher calculation is performed
        devices: list
            A list of the available jax devices (from ``jax.devices()``)
        n_per_device: int
            Number of simulations to handle at once, this should be as large as
            possible without letting the memory overflow for the best
            performance
        validation_fiducial : list of tf.data.Dataset()
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs (for
            validation). These are served ``n_per_device`` at at time as a
            numpy iterator from a TensorFlow dataset.
        validation_derivative : list of tf.data.Dataset()
            The simulations generated at parameter values perturbed from the
            fiducial used to calculate the numerical derivative of network
            outputs with respect to model parameters (for validation).  These
            are served ``n_per_device`` at at time as a numpy iterator from a
            TensorFlow dataset.
        """
        super().__init__(
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser,
            δθ=δθ,
            fiducial=fiducial,
            derivative=derivative,
            validation_fiducial=validation_fiducial,
            validation_derivative=validation_derivative,
            host=host,
            devices=devices,
            n_per_device=n_per_device)

    def _set_data(self, δθ, fiducial, derivative, validation_fiducial,
                  validation_derivative):
        """Checks and sets data attributes with the correct shape

        Parameters
        ----------
        δθ : float(n_params,)
            Size of perturbation to model parameters for the numerical
            derivative
        fiducial : list of tf.data.Datesets
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for fitting)
        derivative : list of tf.data.Datesets
            The derivative of the simulations with respect to the model
            parameters (for fitting)
        validation_fiducial : list of tf.data.Datesets or None, default=None
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for validation). Sets ``validate = True`` attribute if provided
        validation_derivative : list of tf.data.Datesets or None, default=None
            The derivative of the simulations with respect to the model
            parameters (for validation). Sets ``validate = True`` attribute if
            provided

        Raises
        ------
        ValueError
            if δθ is None
        ValueError
            if δθ has wrong shape
        TypeError
            if δθ has wrong type

        Notes
        -----
        No checking is done on the correctness of the tf.data.Dataset
        """
        self.δθ = np.expand_dims(
            _check_input(δθ, (self.n_params,), "δθ"), (0, 1))
        self.fiducial = fiducial
        self.derivative = derivative
        if ((validation_fiducial is not None)
                and (validation_derivative is not None)):
            self.validation_fiducial = validation_fiducial
            self.validation_derivative = validation_derivative
            self.validate = True

    def _set_dataset(self, prefetch=None, cache=None):
        """Overwritten function to prevent building dataset, does list check

        Raises
        ------
        ValueError
            if fiducial or derivative are None
        ValueError
            if any dataset has wrong shape
        TypeError
            if any dataset has wrong type
        """
        _check_type(self.fiducial, list, "fiducial", shape=self.n_devices)
        _check_type(self.derivative, list, "derivative", shape=self.n_devices)
        if self.validate:
            _check_type(self.validation_fiducial, list, "validation_fiducial",
                        shape=self.n_devices)
            _check_type(self.validation_derivative, list,
                        "validation_derivative", shape=self.n_devices)



import pytest
import jax.numpy as np
import jaxlib
import inspect
from test.aggregated_defaults import aggregatedTests
from test.simulator_defaults import simulatorTests
from imnn.imnn import AggregatedSimulatorIMNN


class aggregatedSimulatorTests(aggregatedTests, simulatorTests):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def specific_exceptions(self, variable, input_variable, kwargs):
        if variable == "devices":
            if input_variable is list():
                with pytest.raises(TypeError) as info:
                    self.imnn(**kwargs)
                assert info.match("`devices` has no elements in")
                return True
        if variable == "simulator":
            if callable(input_variable):
                if len(inspect.signature(input_variable).parameters) != 2:
                    kwargs["simulator"] = input_variable
                    with pytest.raises(ValueError) as info:
                        self.imnn(**kwargs)
                    assert info.match(
                        "`simulator` must take two arguments, a JAX prng " +
                        "and simulator parameters.")
                    return True
        if variable == "host":
            if input_variable is None:
                return False
            if not isinstance(input_variable, jaxlib.xla_extension.Device):
                kwargs[variable] = input_variable
                with pytest.raises(TypeError) as info:
                    self.imnn(**kwargs)
                assert info.match(
                    "`host` must be an xla device but is a "
                    f"{type(input_variable)}")
                return True
        return False


test = aggregatedSimulatorTests(
    imnn=AggregatedSimulatorIMNN, filename="aggregated_simulator",
    simulate=True)


def fail_fn(a, b, c):
    pass


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize(
    "input_variable",
    [None, list(), 1., 1, np.zeros((1,)), test.rng, tuple(), (0, 0),
     (test.model[0], 0), test.bad_model, test.state, fail_fn])
@pytest.mark.parametrize("variable", test.kwargs.keys())
def test_initialisation_parameters_(
        variable, kwargs, input_variable, state):
    test.initialise_parameters(
        variable, kwargs, input_variable, state=state, validate=False)


@pytest.mark.parametrize("state", [False, True])
@pytest.mark.parametrize("variable", ["n_s", "n_d", "same"])
def test_splitting_(variable, state):
    test.splitting(variable, test.kwargs, state=state, validate=False)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize(
    "input_variable", [None, list(), 1., 1, np.zeros((1,)), test.rng])
@pytest.mark.parametrize("variable", test.fit_kwargs.keys())
def test_fit_parameters_(variable, kwargs, input_variable, state):
    test.fit_parameters(
        variable, kwargs, test.fit_kwargs, input_variable, state=state,
        validate=False)


@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("none_first", [True, False])
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
def test_combined_running_test_(kwargs, state, validate, fit, none_first):
    test.combined_running_test(
        [test.single_target_data, test.batch_target_data], kwargs,
        test.fit_kwargs, state=state, validate=validate, fit=fit,
        none_first=none_first, aggregated=True)



class container:
    def __init__(self):
        self.all = None
        self.accepted = None
        self.rejected = None
        self.ϵ = None
        self.size = None
        self.n_accepted = None



import jax
import jax.numpy as np
import tensorflow as tf
from functools import partial
from imnn.imnn._aggregated_imnn import _AggregatedIMNN
from imnn.imnn import NumericalGradientIMNN
from imnn.utils import jacrev, add_nested_pytrees
from imnn.utils.utils import _check_boolean, _check_type


class AggregatedNumericalGradientIMNN(_AggregatedIMNN, NumericalGradientIMNN):
    """Information maximising neural network fit using numerical derivatives

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    simulations :math:`{\\bf d}^i` originally generated at fiducial model
    parameter :math:`{\\bf\\theta}^\\rm{fid}`, and a set of
    :math:`i\\in[1, n_d]` simulations,
    :math:`\\{{\\bf d}_{\\alpha^-}^i, {\\bf d}_{\\alpha^+}^i\\}`, generated
    with the same seed at each :math:`i` generated at
    :math:`{\\bf\\theta}^\\rm{fid}` apart from at parameter label
    :math:`\\alpha` with values

    .. math::
        \\theta_{\\alpha^-} = \\theta_\\alpha^\\rm{fid}-\\delta\\theta_\\alpha

    and

    .. math::
        \\theta_{\\alpha^+} = \\theta_\\alpha^\\rm{fid}+\\delta\\theta_\\alpha

    where :math:`\\delta\\theta_\\alpha` is a :math:`n_{params}` length vector
    with the :math:`\\alpha` element having a value which perturbs the
    parameter :math:`\\theta^{\\rm fid}_\\alpha`. This means there are
    :math:`2\\times n_{params}\\times n_d` simulations used to calculate the
    numerical derivatives (this is extremely cheap compared to other machine
    learning methods). All these simulations are passed through a network
    :math:`f_{{\\bf w}}({\\bf d})` with network parameters :math:`{\\bf w}` to
    obtain network outputs :math:`{\\bf x}^i` and
    :math:`\\{{\\bf x}_{\\alpha^-}^i,{\\bf x}_{\\alpha^+}^i\\}`. These
    perturbed values are combined to obtain

    .. math::
        \\frac{\\partial{{\\bf x}^i}}{\\partial\\theta_\\alpha} =
        \\frac{{\\bf x}_{\\alpha^+}^i - {\\bf x}_{\\alpha^-}^i}
        {\\delta\\theta_\\alpha}

    With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    To enable the use of large data (or networks) the whole procedure is
    aggregated. This means that the passing of the simulations through the
    network is farmed out to the desired XLA devices, and recollected,
    ``n_per_device`` inputs at a time. These are then used to calculate the
    automatic gradient of the loss function with respect to the calculated
    summaries and derivatives, :math:`\\partial\\Lambda/\\partial{\\bf x}^i`
    (which is a fairly small computation as long as ``n_summaries`` and ``n_s``
    {and ``n_d``} are not huge). Once this is calculated, the simulations are
    passed through the network AGAIN this time calculating the Jacobian of the
    network output with respect to the network parameters
    :math:`\\partial{\\bf x}^i/\\partial{\\bf w}` which is then combined via
    the chain rule to get

    .. math::
        \\frac{\\partial\\Lambda}{\\partial{\\bf w}} =
        \\frac{\\partial\\Lambda}{\\partial{\\bf x}^i}
        \\frac{\\partial{\\bf x}^i}{\\partial{\\bf w}}

    This can then be passed to the optimiser.

    Parameters
    ----------
    δθ : float(n_params,)
        Size of perturbation to model parameters for the numerical derivative
    fiducial : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for fitting). These
        are served ``n_per_device`` at a time as a numpy iterator from a
        TensorFlow dataset.
    derivative : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network outputs
        with respect to model parameters (for fitting).  These are served
        ``n_per_device`` at a time as a numpy iterator from a TensorFlow
        dataset.
    validation_fiducial : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for validation).
        These are served ``n_per_device`` at a time as a numpy iterator from a
        TensorFlow dataset.
    validation_derivative : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network outputs
        with respect to model parameters (for validation).  These are served
        ``n_per_device`` at a time as a numpy iterator from a TensorFlow
        dataset.
    fiducial_iterations : int
        The number of iterations over the fiducial dataset
    derivative_iterations : int
        The number of iterations over the derivative dataset
    derivative_output_shape : tuple
        The shape of the output of the derivatives from the network
    fiducial_batch_shape : tuple
        The shape of each batch of fiducial simulations (without input or
        summary shape)
    derivative_batch_shape : tuple
        The shape of each batch of derivative simulations (without input or
        summary shape)
    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, fiducial, derivative, δθ,
                 host, devices, n_per_device, validation_fiducial=None,
                 validation_derivative=None, prefetch=None,
                 cache=False):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also fills the
        simulation attributes (and validation if available).

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        fiducial : float(n_s, input_shape)
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for fitting)
        derivative : float(n_d, 2, n_params, input_shape)
            The simulations generated at parameter values perturbed from the
            fiducial used to calculate the numerical derivative of network
            outputs with respect to model parameters (for fitting)
        δθ : float(n_params,)
            Size of perturbation to model parameters for the numerical
            derivative
        host: jax.device
            The main device where the Fisher calculation is performed
        devices: list
            A list of the available jax devices (from ``jax.devices()``)
        n_per_device: int
            Number of simulations to handle at once, this should be as large as
            possible without letting the memory overflow for the best
            performance
        validation_fiducial : float(n_s, input_shape) or None, default=None
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for validation)
        validation_derivative : float(n_d, 2, n_params, input_shape) or None
            The simulations generated at parameter values perturbed from the
            fiducial used to calculate the numerical derivative of network
            outputs with respect to model parameters (for validation)
        prefetch : tf.data.AUTOTUNE or int or None, default=None
            How many simulation to prefetch in the tensorflow dataset
        cache : bool, default=False
            Whether to cache simulations in the tensorflow datasets
        """
        NumericalGradientIMNN.__init__(
            self=self,
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser,
            δθ=δθ,
            fiducial=fiducial,
            derivative=derivative,
            validation_fiducial=validation_fiducial,
            validation_derivative=validation_derivative)
        _AggregatedIMNN.__init__(
            self=self,
            host=host,
            devices=devices,
            n_per_device=n_per_device)
        self._set_dataset(prefetch, cache)

    def _set_shapes(self):
        """Calculates the shapes for batching over different devices
        """
        self.fiducial_iterations = self.n_s \
            // (self.n_devices * self.n_per_device)
        self.derivative_iterations = self.n_d * 2 * self.n_params \
            // (self.n_devices * self.n_per_device)
        self.derivative_output_shape = (
            self.n_d, 2, self.n_params, self.n_summaries)
        self.fiducial_batch_shape = (
            self.n_devices,
            self.fiducial_iterations,
            self.n_per_device)
        self.derivative_batch_shape = (
            self.n_devices,
            self.derivative_iterations,
            self.n_per_device)

    def _set_dataset(self, prefetch, cache):
        """ Transforms the data into lists of tensorflow dataset iterators

        Parameters
        ----------
        prefetch : tf.data.AUTOTUNE or int or None
            How many simulation to prefetch in the tensorflow dataset
        cache : bool
            Whether to cache simulations in the tensorflow datasets

        Raises
        ------
        ValueError
            If ``cache`` and/or ``prefetch`` is None
        TypeError
            If ``cache`` and/or ``prefetch`` is wrong type
        """
        cache = _check_boolean(cache, "cache")
        prefetch = _check_type(prefetch, int, "prefetch", allow_None=True)
        self.fiducial = self.fiducial.reshape(
            self.fiducial_batch_shape + self.input_shape)
        self.fiducial = [
            tf.data.Dataset.from_tensor_slices(
                fiducial)
            for fiducial in self.fiducial]

        self.derivative = self.derivative.reshape(
            self.derivative_batch_shape + self.input_shape)
        self.derivative = [
            tf.data.Dataset.from_tensor_slices(
                derivative)
            for derivative in self.derivative]

        if cache:
            self.fiducial = [
                fiducial.cache()
                for fiducial in self.fiducial]
            self.derivative = [
                derivative.cache()
                for derivative in self.derivative]
        if prefetch is not None:
            self.fiducial = [
                fiducial.prefetch(prefetch)
                for fiducial in self.fiducial]
            self.derivative = [
                derivative.prefetch(prefetch)
                for derivative in self.derivative]

        self.fiducial = [
            fiducial.repeat().as_numpy_iterator()
            for fiducial in self.fiducial]
        self.derivative = [
            derivative.repeat().as_numpy_iterator()
            for derivative in self.derivative]

        if self.validate:
            self.validation_fiducial = self.validation_fiducial.reshape(
                self.fiducial_batch_shape + self.input_shape)
            self.validation_fiducial = [
                tf.data.Dataset.from_tensor_slices(
                    fiducial)
                for fiducial in self.validation_fiducial]

            self.validation_derivative = self.validation_derivative.reshape(
                self.derivative_batch_shape + self.input_shape)
            self.validation_derivative = [
                tf.data.Dataset.from_tensor_slices(
                    derivative)
                for derivative in self.validation_derivative]

            if cache:
                self.validation_fiducial = [
                    fiducial.cache()
                    for fiducial in self.validation_fiducial]
                self.validation_derivative = [
                    derivative.cache()
                    for derivative in self.validation_derivative]
            if prefetch is not None:
                self.validation_fiducial = [
                    fiducial.prefetch(prefetch)
                    for fiducial in self.validation_fiducial]
                self.validation_derivative = [
                    derivative.prefetch(prefetch)
                    for derivative in self.validation_derivative]

            self.validation_fiducial = [
                fiducial.repeat().as_numpy_iterator()
                for fiducial in self.validation_fiducial]
            self.validation_derivative = [
                derivative.repeat().as_numpy_iterator()
                for derivative in self.validation_derivative]

    def _set_batch_functions(self):
        """ Creates jitted functions placed on desired XLA devices

        For each set of summaries to correctly be calculated on a particular
        device we predefine the jitted functions on each of these devices
        """
        self.batch_summaries = [
            jax.jit(
                partial(
                    self._get_batch_summaries,
                    θ=self.θ_fid,
                    gradient=False,
                    derivative=False),
                device=device)
            for device in self.devices]
        self.batch_gradients = [
            jax.jit(
                partial(
                    self._get_batch_summaries,
                    θ=self.θ_fid,
                    gradient=True,
                    derivative=False),
                device=device)
            for device in self.devices]

    def get_summary(self, inputs, w, θ, derivative=False, gradient=False):
        """ Returns a single summary of a simulation or its gradient

        Parameters
        ----------
        inputs : float(input_shape) or tuple
            A single simulation to pass through the network or a tuple of
                - **dΛ_dx** *float(input_shape, n_params)* -- the derivative of
                  the loss function with respect to a network summary
                - **d** *float(input_shape)* -- a simulation to compress with
                  the network
        w : list
            The network parameters
        θ : float(n_params,)
            The value of the parameters to generate the simulation at
            (fiducial), unused if not simulating on the fly
        derivative : bool, default=False
            Whether a derivative of the simulation with respect to model
            parameters is also passed. This must be False for
            NumericalGradientIMNN
        gradient : bool, default=False
            Whether to calculate the gradient with respect to model parameters
        """
        def fn(d, w):
            """ Returns a compressed simulation from the network

            Parameters
            ----------
            d : float(input_shape)
                The simulation to be compressed
            w : list
                The network parameters

            Returns
            ------
            float(n_summaries,):
                The summary from the output of the network
            """
            return self.model(w, d)
        if gradient:
            dΛ_dx, d = inputs
            dx_dw = jacrev(fn, argnums=1)(d, w)
            return self._construct_gradient(dx_dw, aux=dΛ_dx, func="einsum")
        else:
            return fn(inputs, w)

    def _collect_input(self, key, validate=False):
        """ Returns validation or fitting sets

        Parameters
        ----------
        key : None or int(2,)
            Random number generators not used in this case
        validate : bool
            Whether to return the set for validation or for fitting

        Returns
        -------
        list of tf.data.Dataset().as_numpy_iterators:
            The iterators for fiducial simulations for fitting or validation
        list of tf.data.Dataset().as_numpy_iterators:
            The iterators for derivative simulations for fitting or validation
        """
        if validate:
            fiducial = self.validation_fiducial
            derivative = self.validation_derivative
        else:
            fiducial = self.fiducial
            derivative = self.derivative
        return fiducial, derivative

    def get_summaries(self, w, key=None, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        Selects either the fitting or validation sets and loops through the
        iterator on each XLA device to pass them through the network to get the
        network outputs. These are then pushed back to the host for the
        computation of the loss function.

        The fiducial simulations are processed first and then the simulations
        which are varied with respect to model parameters for the derivatives.

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=False
            Whether to get summaries of the validation set

        Returns
        -------
        float(n_s, n_summaries):
            The set of all network outputs used to calculate the covariance
        float(n_d, 2, n_params, n_summaries):
            The outputs of the network of simulations made at perturbed
            parameter values to construct the derivative of the network outputs
            with respect to the model parameters numerically
        """
        d, d_mp = self._collect_input(key, validate=validate)

        x = [np.zeros((self.fiducial_iterations,
                       self.n_per_device,
                       self.n_summaries))
             for i in range(self.n_devices)]
        x_mp = [np.zeros((self.derivative_iterations,
                          self.n_per_device,
                          self.n_summaries))
                for i in range(self.n_devices)]

        for i in range(self.fiducial_iterations):
            for j, (fn, dataset) in enumerate(zip(self.batch_summaries, d)):
                x[j] = jax.ops.index_update(
                    x[j], jax.ops.index[i], fn(next(dataset), w))

        for i in range(self.derivative_iterations):
            for j, (fn, dataset) in enumerate(zip(self.batch_summaries, d_mp)):
                x_mp[j] = jax.ops.index_update(
                    x_mp[j], jax.ops.index[i], fn(next(dataset), w))

        if self.n_devices > 1:
            x = [jax.device_put(_x, device=self.host)
                 for _x in x]
            x_mp = [jax.device_put(_x_mp, device=self.host)
                    for _x_mp in x_mp]

        x = np.stack(x, 0).reshape((self.n_s, self.n_summaries))
        x_mp = np.stack(x_mp, 0).reshape(self.derivative_output_shape)
        return x, x_mp

    def _split_dΛ_dx(self, dΛ_dx):
        """ Returns the gradient of loss function wrt summaries (derivatives)

        The gradient of loss function with respect to network outputs and
        their derivatives with respect to model parameters has to be reshaped
        and aggregated onto each XLA device matching the format that the
        tensorflow dataset feeds simulations.

        Parameters
        ----------
        dΛ_dx : tuple
            - **dΛ_dx** *float(n_s, n_params, n_summaries)* -- the derivative
              of the loss function with respect to network summaries
            - **d2Λ_dxdθ** *float(n_d, 2, n_params, n_summaries)* -- the
              derivative of the loss function with respect to the derivative of
              network summaries with respect to model parameters

        Returns
        -------
        list:
            a list of sets of derivatives of the loss function with respect to
            network summaries placed on each XLA device
        list:
            a list of sets of derivatives of the loss function with respect to
            the derivative of network summaries with respect to model
            parameters
        """
        d2Λ_dxdθ = [jax.device_put(val, device=device) for val, device in zip(
            dΛ_dx[1].reshape(
                self.derivative_batch_shape + (self.n_summaries,)),
            self.devices)]
        dΛ_dx = [jax.device_put(val, device=device) for val, device in zip(
            dΛ_dx[0].reshape(
                self.fiducial_batch_shape + (self.n_summaries,)),
            self.devices)]
        return dΛ_dx, d2Λ_dxdθ

    def get_gradient(self, dΛ_dx, w, key=None):
        """Aggregates gradients together to update the network parameters

        To avoid having to calculate the gradient with respect to all the
        simulations at once we aggregated by addition the gradient calculation
        by looping over the simulations again and combining them with the
        derivative of the loss function with respect to the network outputs
        (and their derivatives with respect to the model parameters). Whilst
        this is expensive, it is necessary since we cannot make a stochastic
        estimate of the Fisher information accurately and therefore we need to
        use all the simulations available - which is probably too large to fit
        in memory.

        Parameters
        ----------
        dΛ_dx : tuple
            - **dΛ_dx** *float(n_s, n_params, n_summaries)* -- the derivative
              of the loss function with respect to network summaries
            - **d2Λ_dxdθ** *float(n_d, 2, n_params, n_summaries)* -- the
              derivative of the loss function with respect to the derivative of
              network summaries with respect to model parameters
        w : list
            Network parameters
        key : None or int(2,)
            Random number generator used in SimulatorIMNN

        Returns
        -------
        list:
            The gradient of the loss function with respect to the network
            parameters calculated by aggregating
        """
        dΛ_dx, dΛ_dx_mp = self._split_dΛ_dx(dΛ_dx)
        d, d_mp = self._collect_input(key, validate=False)

        gradient = [
            jax.jit(
                partial(self._construct_gradient, func="zeros"),
                device=device)(w)
            for device in self.devices]

        for i in range(self.fiducial_iterations):
            for j, (fn, dataset, grad, device) in enumerate(
                    zip(self.batch_gradients, d, dΛ_dx, self.devices)):
                gradient[j] = jax.jit(add_nested_pytrees, device=device)(
                    gradient[j],
                    fn((grad[i], next(dataset)), w))

        for i in range(self.derivative_iterations):
            for j, (fn, dataset, grad, device) in enumerate(
                    zip(self.batch_gradients, d_mp, dΛ_dx_mp, self.devices)):
                gradient[j] = jax.jit(add_nested_pytrees, device=device)(
                    gradient[j],
                    fn((grad[i], next(dataset)), w))

        return list(jax.jit(add_nested_pytrees, device=self.host)(*gradient))



import jax
import jax.numpy as np
import tqdm.auto as tqdm
from jax.tree_util import tree_flatten, tree_unflatten
from functools import partial
from imnn.utils.utils import _check_host, _check_devices, _check_boolean, \
    _check_type, _check_splitting, _check_input


class _AggregatedIMNN:
    """Manual aggregation of gradients for the IMNN parent class

    This class defines the overriding fitting functions for
    :func:`~imnn.imnn._imnn._IMNN` which allows gradients to be aggregated
    manually. This is necessary if networks or input data are extremely large
    (or if the number of simulations necessary to estimate the covariance of
    network outputs, ``n_s``, is very large) since all operations may not fit
    in memory.

    The aggregation is done by calculating ``n_per_device`` network outputs at
    once on each available ``jax.device()`` and then scanning over all ``n_s``
    inputs and ``n_d`` simulations necessary to calculate the derivative of the
    mean of the network outputs with respect to the model parameters. This
    gives a set of summaries and derivatives from which the loss function

    .. math::
        \\Lambda = -\\log|\\bf{F}| + r(\\Lambda_2) \\Lambda_2

    (See :doc:`/pages/details`) can be calculated and its gradient with respect
    to these summaries, :math:`\\frac{\\partial\\Lambda}{\\partial x_i^j}` and
    derivatives :math:`\\frac{\\partial\\Lambda}{\\partial\\partial{x_i^j}/
    \\partial\\theta_\\alpha}` calculated, where :math:`i` labels the network
    output and :math:`j` labels the simulation. Note that these are small in
    comparison to the gradient with respect to the network parameters since
    their sizes are ``n_s * n_summaries`` and ``n_d * n_summaries * n_params``
    respectively. Once :math:`\\frac{\\partial\\Lambda}{\\partial{x_i^j}}` and
    :math:`\\frac{\\partial\\Lambda}{\\partial\\partial{x_i^j}/\\partial
    \\theta_\\alpha}` are calculated then gradients of the network outputs with
    respect to the network parameters
    :math:`\\frac{\\partial{x_i^j}}{\\partial{w_{ab}^l}}` and
    :math:`\\frac{\\partial\\partial{x_i^j}/\\partial\\theta_\\alpha}
    {\\partial{w_{ab}^l}}` are calculated and the chain rule is used to get

    .. math::
        \\frac{\\partial\\Lambda}{\\partial{w_{ab}^l}} = \\frac{\\partial
        \\Lambda}{\\partial{x_i^j}}\\frac{\\partial{x_i^j}}
        {\\partial{w_{ab}^l}} + \\frac{\\partial\\Lambda}
        {\\partial\\partial{x_i^j}/\\partial\\theta_\\alpha}
        \\frac{\\partial\\partial{x_i^j}/\\partial\\theta_\\alpha}
        {\\partial{w_{ab}^l}}

    Note that we keep the memory use low because only ``n_per_device``
    simulations are handled at once before being summed into a single gradient
    list on each device.

    ``n_per_devices`` should be as large as possible to get the best
    performance. If everything will fit in memory then this class should be
    avoided.

    The AttributedIMNN class doesn't directly inherit from
    :func:`~imnn.imnn._imnn._IMNN`, but is meant to be built within a child
    class of it. For this reason there are attributes which are not explicitly
    set here, but are used within the module. These will be noted in
    Other Parameters below.

    Parameters
    ----------
    host: jax.device
        The main device where the Fisher information calculation is performed
    devices: list
        A list of the available jax devices (from ``jax.devices()``)
    n_devices: int
        Number of devices to aggregated calculation over
    n_per_device: int
        Number of simulations to handle at once, this should be as large as
        possible without letting the memory overflow for the best performance

    Methods
    -------
    model:
        Neural network as a function of network parameters and inputs
    _get_parameters:
        Function which extracts the network parameters from the state
    _model_initialiser:
        Function to initialise neural network weights from RNG and shape tuple
    _opt_initialiser:
        Function which generates the optimiser state from network parameters
    _update:
        Function which updates the state from a gradient
    batch_summaries:
        Jitted function to calculate summaries on each XLA device
    batch_summaries_with_derivatives:
        Jitted function to calculate summaries from derivative on each device
    batch_gradients:
        Jitted function to calculate gradient on each XLA device
    batch_gradients_with_derivatives:
        Jitted function to calculate gradient from derivative on eachdevice
    """
    def __init__(self, host, devices, n_per_device):
        """Constructor method

        Parameters
        ----------
        host: jax.device
            The main device where the Fisher calculation is performed
        devices: list
            A list of the available jax devices (from ``jax.devices()``)
        n_per_device: int
            Number of simulations to handle at once, this should be as large as
            possible without letting the memory overflow for the best
            performance
        """
        self.host = _check_host(host)
        self._set_devices(devices, n_per_device)
        self._set_batch_functions()
        self._set_shapes()

    def _set_devices(self, devices, n_per_device):
        """Checks that devices exist and that reshaping onto devices can occur

        Due to the aggregation then balanced splits must be made between the
        different devices and so these are checked.

        Parameters
        ----------
        devices: list
            A list of the available jax devices (from ``jax.devices()``)
        n_per_device: int
            Number of simulations to handle at once, this should be as large as
            possible without letting the memory overflow for the best
            performance

        Raises
        ------
        ValueError
            If ``devices`` or ``n_per_device`` are None
        ValueError
            If balanced splitting cannot be done
        TypeError
            If ``devices`` is not a list and if ``n_per_device`` is not an int
        """
        self.devices = _check_devices(devices)
        self.n_devices = len(self.devices)
        self.n_per_device = _check_type(n_per_device, int, "n_per_device")
        if self.n_s == self.n_d:
            _check_splitting(self.n_s, "n_s and n_d", self.n_devices,
                             self.n_per_device)
        else:
            _check_splitting(
                self.n_s, "n_s", self.n_devices, self.n_per_device)
            _check_splitting(
                self.n_d, "n_d", self.n_devices, self.n_per_device)

    def _set_batch_functions(self):
        """ Creates jitted functions placed on desired XLA devices

        For each set of summaries to correctly be calculated on a particular
        device we predefine the jitted functions on each of these devices
        """
        self.batch_summaries = [
            jax.jit(
                partial(
                    self._get_batch_summaries,
                    θ=self.θ_fid,
                    gradient=False,
                    derivative=False),
                device=device)
            for device in self.devices]
        self.batch_summaries_with_derivatives = [
            jax.jit(
                partial(
                    self._get_batch_summaries,
                    θ=self.θ_fid,
                    gradient=False,
                    derivative=True),
                device=device)
            for device in self.devices]
        self.batch_gradients = [
            jax.jit(
                partial(
                    self._get_batch_summaries,
                    θ=self.θ_fid,
                    gradient=True,
                    derivative=False),
                device=device)
            for device in self.devices]
        self.batch_gradients_with_derivatives = [
            jax.jit(
                partial(
                    self._get_batch_summaries,
                    θ=self.θ_fid,
                    gradient=True,
                    derivative=True),
                device=device)
            for device in self.devices]

    def _set_shapes(self):
        """Calculates the shapes for batching over different devices

        Not implemented

        Raises
        ------
        ValueError
            Not implemented in _AggregatedIMNN
        """
        raise ValueError("`_set_shapes` not implemented in `_AggregatedIMNN`")

    def fit(self, λ, ϵ, rng=None, patience=100, min_iterations=100,
            max_iterations=int(1e5), print_rate=None, best=True):
        """Fitting routine for the IMNN

        Parameters
        ----------
        λ : float
            Coupling strength of the regularisation
        ϵ : float
            Closeness criterion describing how close to the 1 the determinant
            of the covariance (and inverse covariance) of the network outputs
            is desired to be
        rng : int(2,) or None, default=None
            Stateless random number generator
        patience : int, default=10
            Number of iterations where there is no increase in the value of the
            determinant of the Fisher information matrix, used for early
            stopping
        min_iterations : int, default=100
            Number of iterations that should be run before considering early
            stopping using the patience counter
        max_iterations : int, default=int(1e5)
            Maximum number of iterations to run the fitting procedure for
        print_rate : int or None, default=None,
            Number of iterations before updating the progress bar whilst
            fitting. There is a performance hit from updating the progress bar
            more often and there is a large performance hit from using the
            progress bar at all. (Possible ``RET_CHECK`` failure if
            ``print_rate`` is not ``None`` when using GPUs).
            For this reason it is set to None as default
        best : bool, default=True
            Whether to set the network parameter attribute ``self.w`` to the
            parameter values that obtained the maximum determinant of
            the Fisher information matrix or the parameter values at the final
            iteration of fitting

        Example
        -------

        We are going to summarise the mean and variance of some random Gaussian
        noise with 10 data points per example using an AggregatedSimulatorIMNN.
        In this case we are going to generate the simulations on-the-fly with a
        simulator written in jax (from the examples directory). These
        simulations will be generated on-the-fly and passed through the network
        on each of the GPUs in ``jax.devices("gpu")`` and we will make 100
        simulations on each device at a time. The main computation will be done
        on the CPU. We will use 1000 simulations to estimate the covariance of
        the network outputs and the derivative of the mean of the network
        outputs with respect to the model parameters (Gaussian mean and
        variance) and generate the simulations at a fiducial μ=0 and Σ=1. The
        network will be a stax model with hidden layers of ``[128, 128, 128]``
        activated with leaky relu and outputting 2 summaries. Optimisation will
        be via Adam with a step size of ``1e-3``. Rather arbitrarily we'll set
        the regularisation strength and covariance identity constraint to λ=10
        and ϵ=0.1 (these are relatively unimportant for such an easy model).

        .. code-block:: python

            import jax
            import jax.numpy as np
            from jax.experimental import stax, optimizers
            from imnn import AggregatedSimulatorIMNN

            rng = jax.random.PRNGKey(0)

            n_s = 1000
            n_d = 1000
            n_params = 2
            n_summaries = 2
            input_shape = (10,)
            θ_fid = np.array([0., 1.])

            def simulator(rng, θ):
                return θ[0] + jax.random.normal(
                    rng, shape=input_shape) * np.sqrt(θ[1])

            model = stax.serial(
                stax.Dense(128),
                stax.LeakyRelu,
                stax.Dense(128),
                stax.LeakyRelu,
                stax.Dense(128),
                stax.LeakyRelu,
                stax.Dense(n_summaries))
            optimiser = optimizers.adam(step_size=1e-3)

            λ = 10.
            ϵ = 0.1

            model_key, fit_key = jax.random.split(rng)

            host = jax.devices("cpu")[0]
            devices = jax.devices("gpu")

            n_per_device = 100

            imnn = AggregatedSimulatorIMNN(
                n_s=n_s, n_d=n_d, n_params=n_params, n_summaries=n_summaries,
                input_shape=input_shape, θ_fid=θ_fid, model=model,
                optimiser=optimiser, key_or_state=model_key,
                simulator=simulator, host=host, devices=devices,
                n_per_device=n_per_device)

            imnn.fit(λ, ϵ, rng=fit_key, min_iterations=1000, patience=250,
                     print_rate=None)


        Notes
        -----
        A minimum number of interations should be be run before stopping based
        on a maximum determinant of the Fisher information achieved since the
        loss function has dual objectives. Since the determinant of the
        covariance of the network outputs is forced to 1 quickly, this can be
        at the detriment to the value of the determinant of the Fisher
        information matrix early in the fitting procedure. For this reason
        starting early stopping after the covariance has converged is advised.
        This is not currently implemented but could be considered in the
        future.

        The best fit network parameter values are probably not the most
        representative set of parameters when simulating on-the-fly since there
        is a high chance of a statistically overly-informative set of data
        being generated. Instead, if using
        :func:`~imnn.AggregatedSimulatorIMNN.fit()` consider using
        ``best=False`` which sets ``self.w=self.final_w`` which are the network
        parameter values obtained in the last iteration. Also consider using a
        larger ``patience`` value if using :func:`~imnn.SimulatorIMNN.fit()`
        to overcome the fact that a flukish high value for the determinant
        might have been obtained due to the realisation of the dataset.

        Raises
        ------
        TypeError
            If any input has the wrong type
        ValueError
            If any input (except ``rng``) are ``None``
        ValueError
            If ``rng`` has the wrong shape
        ValueError
            If ``rng`` is ``None`` but simulating on-the-fly

        Methods
        -------
        get_keys_and_params:
            Jitted collection of parameters and random numbers
        calculate_loss:
            Returns the jitted gradient of the loss function wrt summaries
        validation_loss:
            Jitted loss and auxillary statistics from validation set

        Todo
        ----
        - ``rng`` is currently only used for on-the-fly simulation but could
          easily be updated to allow for stochastic models
        - Automatic detection of convergence based on value ``r`` when early
          stopping can be started
        """
        @jax.jit
        def get_keys_and_params(rng, state):
            """Jitted collection of parameters and random numbers

            Parameters
            ----------
            rng : int(2,) or None, default=None
                Stateless random number generator
            state : :obj:state
                The optimiser state used for updating the network parameters
                and optimisation algorithm

            Returns
            -------
            int(2,) or None, default=None:
                Stateless random number generator
            int(2,) or None, default=None:
                Stateless random number generator for training
            int(2,) or None, default=None:
                Stateless random number generator for validation
            list:
                Network parameter values
            """
            rng, training_key, validation_key = self._get_fitting_keys(rng)
            w = self._get_parameters(state)
            return rng, training_key, validation_key, w

        @jax.jit
        @partial(jax.grad, argnums=(0, 1), has_aux=True)
        def calculate_loss(summaries, summary_derivatives):
            """Returns the jitted gradient of the loss function wrt summaries

            Used to calculate the gradient of the loss function wrt summaries
            and derivatives of the summaries with respect to model parameters
            which will be used to calculate the aggregated gradient of the
            Fisher information with respect to the network parameters via the
            chain rule.

            Parameters
            ----------
            summaries : float(n_s, n_summaries)
                The network outputs
            summary_derivatives : float(n_d, n_summaries, n_params)
                The derivative of the network outputs wrt the model parameters

            Returns
            -------
            tuple:
                Gradient of the loss function with respect to network outputs
                and their derivatives with respect to physical model parameters
            tuple:
                Fitting statistics calculated on a single iteration
                    - **F** *(float(n_params, n_params))* -- Fisher information
                      matrix
                    - **C** *(float(n_summaries, n_summaries))* -- covariance
                      of network outputs
                    - **invC** *(float(n_summaries, n_summaries))* -- inverse
                      covariance of network outputs
                    - **Λ2** *(float)* -- covariance regularisation
                    - **r** *(float)* -- regularisation coupling strength
            """
            return self._calculate_loss(
                summaries, summary_derivatives, λ, α)

        @jax.jit
        def validation_loss(summaries, derivatives):
            """Jitted loss and auxillary statistics from validation set

            Parameters
            ----------
            summaries : float(n_s, n_summaries)
                The network outputs
            summary_derivatives : float(n_d, n_summaries, n_params)
                The derivative of the network outputs wrt the model parameters

            Returns
            -------
            tuple:
                Fitting statistics calculated on a single validation iteration
                    - **F** *(float(n_params, n_params))* -- Fisher information
                      matrix
                    - **C** *(float(n_summaries, n_summaries))* -- covariance
                      of network outputs
                    - **invC** *(float(n_summaries, n_summaries))* -- inverse
                      covariance of network outputs
                    - **Λ2** *(float)* -- covariance regularisation
                    - **r** *(float)* -- regularisation coupling strength
            """
            F, C, invC, *_ = self._calculate_F_statistics(
                summaries, derivatives)
            _Λ2 = self._get_regularisation(C, invC)
            _r = self._get_regularisation_strength(_Λ2, λ, α)
            return (F, C, invC, _Λ2, _r)

        λ = _check_type(λ, float, "λ")
        ϵ = _check_type(ϵ, float, "ϵ")
        α = self.get_α(λ, ϵ)
        patience = _check_type(patience, int, "patience")
        min_iterations = _check_type(min_iterations, int, "min_iterations")
        max_iterations = _check_type(max_iterations, int, "max_iterations")
        best = _check_boolean(best, "best")
        if self.simulate and (rng is None):
            raise ValueError("`rng` is necessary when simulating.")
        rng = _check_input(rng, (2,), "rng", allow_None=True)
        max_detF, best_w, detF, detC, detinvC, Λ2, r, counter, \
            patience_counter, state, rng = self._set_inputs(
                rng, max_iterations)
        pbar, print_rate, remainder = self._setup_progress_bar(
            print_rate, max_iterations)
        while self._fit_cond(
                (max_detF, best_w, detF, detC, detinvC, Λ2, r, counter,
                 patience_counter, state, rng),
                patience=patience, max_iterations=max_iterations):
            rng, training_key, validation_key, w = get_keys_and_params(
                rng, state)
            summaries, summary_derivatives = self.get_summaries(
                w=w, key=training_key)
            dΛ_dx, results = calculate_loss(summaries, summary_derivatives)
            grad = self.get_gradient(dΛ_dx, w, key=training_key)
            state = self._update(counter, grad, state)
            w = self._get_parameters(state)
            detF, detC, detinvC, Λ2, r = self._update_history(
                results, (detF, detC, detinvC, Λ2, r), counter, 0)
            if self.validate:
                summaries, summary_derivatives = self.get_summaries(
                    w=w, key=training_key, validate=True)
                results = validation_loss(summaries, summary_derivatives)
                detF, detC, detinvC, Λ2, r = self._update_history(
                    results, (detF, detC, detinvC, Λ2, r), counter, 1)
            _detF = np.linalg.det(results[0])
            patience_counter, counter, _, max_detF, __, best_w = \
                jax.lax.cond(
                    np.greater(_detF, max_detF),
                    self._update_loop_vars,
                    lambda inputs: self._check_loop_vars(
                        inputs, min_iterations),
                    (patience_counter, counter, _detF, max_detF, w, best_w))
            self._update_progress_bar(
                pbar, counter, patience_counter, max_detF, detF[counter],
                detC[counter], detinvC[counter], Λ2[counter], r[counter],
                print_rate, max_iterations, remainder)
            counter += 1
        self._update_progress_bar(
            pbar, counter, patience_counter, max_detF, detF[counter - 1],
            detC[counter - 1], detinvC[counter - 1], Λ2[counter - 1],
            r[counter - 1], print_rate, max_iterations, remainder, close=True)
        self.history["max_detF"] = max_detF
        self.best_w = best_w
        self._set_history(
            (detF[:counter],
             detC[:counter],
             detinvC[:counter],
             Λ2[:counter],
             r[:counter]))
        self.state = state
        self.final_w = self._get_parameters(self.state)
        if best:
            w = self.best_w
        else:
            w = self.final_w
        self.set_F_statistics(w, key=rng)

    def _setup_progress_bar(self, print_rate, max_iterations):
        """Construct progress bar

        Parameters
        ----------
        print_rate : int or None
            The rate at which the progress bar is updated (no bar if None)
        max_iterations : int
            The maximum number of iterations, used to setup bar upper limit

        Returns
        -------
        progress bar or None:
            The TQDM progress bar object
        int or None:
            The print rate (after checking for int or None)
        int or None:
            The difference between the max_iterations and the print rate

        Raises
        ------
        TypeError:
            If ``print_rate`` is not an integer
        """
        print_rate = _check_type(
            print_rate, int, "print_rate", allow_None=True)
        if print_rate is not None:
            if max_iterations < 10000:
                pbar = tqdm.tqdm(total=max_iterations)
            else:
                pbar = tqdm.tqdm()
            remainder = max_iterations % print_rate
            return pbar, print_rate, remainder
        else:
            return None, None, None

    def _update_progress_bar(
            self, pbar, counter, patience_counter, max_detF, detF, detC,
            detinvC, Λ2, r, print_rate, max_iterations, remainder,
            close=False):
        """Updates (and closes) progress bar

        Checks whether a pbar is used and is so checks whether the iteration
        coincides with the print rate, or is the last set of iterations within
        the print rate from the last iteration, or if the last iteration has
        been reached and the bar should be closed.

        Parameters
        ----------
        pbar : progress bar object
            The TQDM progress bar
        counter : int
            The value of the current iteration
        patience_counter : int
            The number of iterations where the maximum of the determinant of
            the Fisher information matrix has not increased
        max_detF : float
            Maximum of the determinant of the Fisher information matrix
        detF : float(n_params, n_params)
            Fisher information matrix
        detC : float(n_summaries, n_summaries)
            Covariance of the network summaries
        detinvC : float(n_summaries, n_summaries)
            Inverse covariance of the network summaries
        Λ2 : float
            Value of the regularisation term
        r : float
            Value of the dynamic regularisation coupling strength
        print_rate : int or None
            The number of iterations to run before updating the progress bar
        max_iterations : int
            The maximum number of iterations to run
        remainder : int or None
            The number of iterations before max_iterations to check progress
        close : bool, default=False
            Whether to close the progress bar (on final iteration)
        """
        if print_rate is not None:
            if ((counter % print_rate == 0)
                    and (counter != max_iterations - remainder)) \
                    or (counter == max_iterations - remainder) \
                    or close:
                postfix_dict = dict(
                    patience=patience_counter, detF=detF, detC=detC,
                    detinvC=detinvC, Λ2=Λ2, r=r)
                if (counter != max_iterations - remainder):
                    pbar.update(print_rate)
                elif close:
                    pbar.update(int(counter - pbar.n))
                    postfix_dict["max_detF"] = max_detF
                else:
                    pbar.update(remainder)
                pbar.set_postfix(postfix_dict)
            if close:
                pbar.close()

    def _collect_input(self, key, validate=False):
        """Returns the dataset to be interated over

        Parameters
        ----------
        key: int(2,)
            A random number generator
        validate: bool, default=False
            Whether to use the validation set or not
        """
        raise ValueError("`collect_input` not implemented")

    def _get_batch_summaries(
            self, inputs, w, θ, gradient=False, derivative=False):
        """Vectorised batch calculation of summaries or gradients

        Parameters
        ----------
        inputs: tuple
            - **dΛ_dx** if ``gradient`` *(float(n_per_device, n_summaries) or
              tuple)*

                - **dΛ_dx** *(float(n_per_device, n_summaries))* -- The
                  gradient of the loss function with respect to network outputs
                - **d2Λ_dxdθ** if ``derivative``
                  *(float(n_per_device, n_summaries, n_params))* -- The
                  gradient of the loss function with respect to derivative of
                  network outputs with respect to model parameters
            - **keys** if :func:`~imnn.SimulatorIMNN`
              *(int(n_per_device, 2))* -- The keys for generating simulations
              on-the-fly
            - **d** if :func:`~imnn.NumericalGradientIMNN`
              *(float(n_per_device, input_shape) or tuple)*

                - **d** *(float(n_per_device, input_shape))* -- The simulations
                  to be evaluated
                - **dd_dθ** if ``derivative``
                  *(float(n_per_device, input_shape, n_params))* -- The
                  derivative of the simulations to be evaluated with respect to
                  model parameters
        w: list
            Network model parameters
        θ: float(n_params,)
            The value of the model parameters to generate simulations at/to
            perform the derivative calculation
        gradient: bool
            Whether to do the gradient calculation
        derivative: bool, default=False
            Whether the gradient of loss function with respect to the
            derivative of the network outputs with respect to the model
            parameters is being used

        Returns
        -------
        float(n_devices, n_per_device, n_summaries) or tuple:
            - **x** if not ``gradient`` *(float(n_per_device, n_summaries) or
              tuple)*

                - **x** *(float(n_per_device, n_summaries))* -- The network
                  outputs
                - **dd_dθ** if ``derivative``
                  *(float(n_per_device, n_summaries, n_params))* -- The
                  derivative of the network outputs with respect to model
                  parameters
            - if ``gradient``
                *(tuple)* -- The accumlated and aggregated gradient of the loss
                function with respect to the network parameters
        """
        @jax.vmap
        def fn(inputs):
            return self.get_summary(
                inputs, w, θ, derivative=derivative, gradient=gradient)
        result = fn(inputs)
        if gradient:
            result = self._construct_gradient(result, func="sum")
        return result

    def get_summaries(self, w, key=None, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        Parameters
        ----------
        w: list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key: int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate: bool, default=True
            Whether to get summaries of the validation set

        Returns
        -------
        float(n_s, n_summaries)
            The network outputs
        float(n_d, n_summaries, n_params)
            The derivative of the network outputs wrt the model parameters

        Raises
        ------
        ValueError
            function not implemented in parent class
        """
        raise ValueError(
            "``get_summaries`` not implemented in ``_AggregatedIMNN``")

    def _split_dΛ_dx(self, dΛ_dx):
        """Separates dΛ_dx and d2Λ_dxdθ and reshapes them for aggregation

        Parameters
        ----------
        dΛ_dx: tuple
            - **dΛ_dx** *(float(n_s, n_summaries))* -- The derivative of the
              loss function wrt the network outputs
            - **d2Λ_dxdθ** *(float(n_d, n_summaries, n_params))* -- The
              derivative of the loss function wrt the derivative of the network
              outputs wrt the model parameters

        Raises
        ------
        ValueError
            function not implemented in parent class
        """
        raise ValueError(
            "``_split_dΛ_dx`` not implemented in ``_AggregatedIMNN``")

    def get_gradient(self, dΛ_dx, w, key=None):
        """Aggregates gradients together to update the network parameters

        To avoid having to calculate the gradient with respect to all the
        simulations at once we aggregate by addition the gradient calculation
        by looping over the simulations again and combining them with the
        derivative of the loss function with respect to the network outputs
        (and their derivatives with respect to the model parameters). Whilst
        this is expensive, it is necessary since we cannot make a stochastic
        estimate of the Fisher information accurately and therefore we need to
        use all the simulations available - which is probably too large to fit
        in memory.

        Parameters
        ----------
        dΛ_dx : tuple
            - **dΛ_dx** *float(n_s, n_params, n_summaries)* -- the derivative
              of the loss function with respect to network summaries
            - **d2Λ_dxdθ** *float(n_d, n_summaries, n_params)* -- the
              derivative of the loss function with respect to the derivative of
              network summaries with respect to model parameters
        w : list
            Network parameters
        key : None or int(2,)
            Random number generator used in SimulatorIMNN

        Returns
        -------
        list:
            The gradient of the loss function with respect to the network
            parameters calculated by aggregating

        Raises
        ------
        ValueError
            function not implemented in parent class
        """
        raise ValueError(
            "``get_gradient`` not implemented in ``_AggregatedIMNN``")

    def _construct_gradient(self, layers, aux=None, func="zeros"):
        """Multiuse function to iterate over tuple of network parameters

        The options are:
            - ``"zeros"`` -- to create an empty gradient array
            - ``"einsum"`` -- to combine tuple of ``dx_dw`` with ``dΛ_dx``
            - ``"derivative_einsum"`` -- to combine tuple of ``d2x_dwdθ`` with
              ``d2Λ_dxdθ``
            - ``"sum"`` -- to reduce sum batches of gradients on the first axis

        Parameters
        ----------
        layers: tuple
            The tuple of tuples of arrays to be iterated over
        aux: float(various shapes)
            parameter to pass dΛ_dx and d2Λ_dxdθ to einsum
        func: str
            Option for the function to apply
                - ``"zeros"`` -- to create an empty gradient array
                - ``"einsum"`` -- to combine tuple of ``dx_dw`` with ``dΛ_dx``
                - ``"derivative_einsum"`` -- to combine tuple of ``d2x_dwdθ``
                  with ``d2Λ_dxdθ``
                - ``"sum"`` -- to reduce sum batches of gradients on the first
                  axis

        Returns
        -------
        tuple:
            Tuple of objects like the gradient of the loss function with
            respect to the network parameters

        Raises
        ------
        ValueError
            If applied function is not implemented
        """
        if func == "zeros":
            def func(x):
                return np.zeros_like(x)
        elif func == "einsum":
            def func(x):
                return np.einsum("i,i...->...", aux, x)
        elif func == "derivative_einsum":
            def func(x):
                return np.einsum("ij,ij...->...", aux, x)
        elif func == "sum":
            def func(x):
                return np.sum(x, 0)
        else:
            raise ValueError("`func` must be `'zeros'`, `'einsum'`, " +
                             "`'derivative_einsum'` or `'sum'`")

        value_flat, value_tree = tree_flatten(layers)
        transformed_flat = list(map(func, value_flat))
        return tree_unflatten(value_tree, transformed_flat)



import pytest
import jax.numpy as np
import tensorflow as tf
from test.aggregated_defaults import aggregatedTests
from test.numerical_gradient_defaults import numericalGradientTests
from imnn.imnn import AggregatedNumericalGradientIMNN


class aggregatedNumericalGradientTests(
        aggregatedTests, numericalGradientTests):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.kwargs["prefetch"] = tf.data.AUTOTUNE
        self.reduced_kwargs["prefetch"] = tf.data.AUTOTUNE
        self.kwargs["cache"] = True
        self.reduced_kwargs["cache"] = True


test = aggregatedNumericalGradientTests(
    imnn=AggregatedNumericalGradientIMNN,
    filename="aggregated_numerical_gradient")


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable",
    [None, list(), 1., 1, np.zeros((1,)), test.rng, tuple(), (0, 0),
     (test.model[0], 0), test.bad_model, test.state])
@pytest.mark.parametrize("variable", test.kwargs.keys())
def test_initialisation_parameters_(
        variable, kwargs, input_variable, state, validate):
    test.initialise_parameters(
        variable, kwargs, input_variable, state=state, validate=validate)


@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("state", [False, True])
@pytest.mark.parametrize("variable", ["n_s", "n_d", "same"])
def test_splitting_(variable, validate, state):
    test.splitting(variable, test.kwargs, state=state, validate=validate)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable", [None, list(), 1., 1, np.zeros((1,)), test.rng])
@pytest.mark.parametrize("variable", test.fit_kwargs.keys())
def test_fit_parameters_(variable, kwargs, input_variable, state, validate):
    test.fit_parameters(
        variable, kwargs, test.fit_kwargs, input_variable, state=state,
        validate=validate)


@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("none_first", [True, False])
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
def test_combined_running_test_(kwargs, state, validate, fit, none_first):
    test.combined_running_test(
        [test.single_target_data, test.batch_target_data], kwargs,
        test.fit_kwargs, state=state, validate=validate, fit=fit,
        none_first=none_first, aggregated=True)



import jax.numpy as np
from imnn.imnn._imnn import _IMNN
from imnn.utils.utils import _check_input


class NumericalGradientIMNN(_IMNN):
    """Information maximising neural network fit using numerical derivatives

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    simulations :math:`{\\bf d}^i` originally generated at fiducial model
    parameter :math:`{\\bf\\theta}^\\rm{fid}`, and a set of
    :math:`i\\in[1, n_d]` simulations,
    :math:`\\{{\\bf d}_{\\alpha^-}^i, {\\bf d}_{\\alpha^+}^i\\}`, generated
    with the same seed at each :math:`i` generated at
    :math:`{\\bf\\theta}^\\rm{fid}` apart from at parameter label
    :math:`\\alpha` with values

    .. math::
        \\theta_{\\alpha^-} = \\theta_\\alpha^\\rm{fid}-\\delta\\theta_\\alpha

    and

    .. math::
        \\theta_{\\alpha^+} = \\theta_\\alpha^\\rm{fid}+\\delta\\theta_\\alpha

    where :math:`\\delta\\theta_\\alpha` is a :math:`n_{params}` length vector
    with the :math:`\\alpha` element having a value which perturbs the
    parameter :math:`\\theta^{\\rm fid}_\\alpha`. This means there are
    :math:`2\\times n_{params}\\times n_d` simulations used to calculate the
    numerical derivatives (this is extremely cheap compared to other machine
    learning methods). All these simulations are passed through a network
    :math:`f_{{\\bf w}}({\\bf d})` with network parameters :math:`{\\bf w}` to
    obtain network outputs :math:`{\\bf x}^i` and
    :math:`\\{{\\bf x}_{\\alpha^-}^i,{\\bf x}_{\\alpha^+}^i\\}`. These
    perturbed values are combined to obtain

    .. math::
        \\frac{\\partial{{\\bf x}^i}}{\\partial\\theta_\\alpha} =
        \\frac{{\\bf x}_{\\alpha^+}^i - {\\bf x}_{\\alpha^-}^i}
        {\\delta\\theta_\\alpha}

    With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    Once the loss function is calculated the automatic gradient is then
    calculated and used to update the network parameters via the optimiser
    function.

    Parameters
    ----------
    δθ : float(n_params,)
        Size of perturbation to model parameters for the numerical derivative
    fiducial : float(n_s, input_shape)
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for fitting)
    derivative : float(n_d, 2, n_params, input_shape)
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network outputs
        with respect to model parameters (for fitting)
    validation_fiducial : float(n_s, input_shape) or None
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for validation)
    validation_derivative : float(n_d, 2, n_params, input_shape) or None
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network outputs
        with respect to model parameters (for validation)
    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, fiducial, derivative, δθ,
                 validation_fiducial=None, validation_derivative=None):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also fills the
        simulation attributes (and validation if available).

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        fiducial : float(n_s, input_shape)
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for fitting)
        derivative : float(n_d, 2, n_params, input_shape)
            The simulations generated at parameter values perturbed from the
            fiducial used to calculate the numerical derivative of network
            outputs with respect to model parameters (for fitting)
        δθ : float(n_params,)
            Size of perturbation to model parameters for the numerical
            derivative
        validation_fiducial : float(n_s, input_shape) or None, default=None
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for validation)
        validation_derivative : float(n_d, 2, n_params, input_shape) or None
            The simulations generated at parameter values perturbed from the
            fiducial used to calculate the numerical derivative of network
            outputs with respect to model parameters (for validation)
        """
        super().__init__(
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser)
        self._set_data(δθ, fiducial, derivative, validation_fiducial,
                       validation_derivative)

    def _set_data(self, δθ, fiducial, derivative, validation_fiducial,
                  validation_derivative):
        """Checks and sets data attributes with the correct shape

        Parameters
        ----------
        δθ : float(n_params,)
            Size of perturbation to model parameters for the numerical
            derivative
        fiducial : float(n_s, input_shape)
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for fitting)
        derivative : float(n_d, input_shape, n_params)
            The derivative of the simulations with respect to the model
            parameters (for fitting)
        validation_fiducial : float(n_s, input_shape) or None, default=None
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for validation). Sets ``validate = True`` attribute if provided
        validation_derivative : float(n_d, input_shape, n_params) or None
            The derivative of the simulations with respect to the model
            parameters (for validation). Sets ``validate = True`` attribute if
            provided
        """
        self.δθ = np.expand_dims(
            _check_input(δθ, (self.n_params,), "δθ"), (0, 1))
        self.fiducial = _check_input(
            fiducial, (self.n_s,) + self.input_shape, "fiducial")
        self.derivative = _check_input(
            derivative, (self.n_d, 2, self.n_params) + self.input_shape,
            "derivative")
        if ((validation_fiducial is not None)
                and (validation_derivative is not None)):
            self.validation_fiducial = _check_input(
                validation_fiducial, (self.n_s,) + self.input_shape,
                "validation_fiducial")
            self.validation_derivative = _check_input(
                validation_derivative,
                (self.n_d, 2, self.n_params) + self.input_shape,
                "validation_derivative")
            self.validate = True

    def _collect_input(self, key, validate=False):
        """ Returns validation or fitting sets

        Parameters
        ----------
        key : None or int(2,)
            Random number generators not used in this case
        validate : bool
            Whether to return the set for validation or for fitting

        Returns
        -------
        float(n_s, input_shape):
            The fiducial simulations for fitting or validation
        float(n_d, 2, n_params, input_shape):
            The derivative simulations for fitting or validation
        """
        if validate:
            fiducial = self.validation_fiducial
            derivative = self.validation_derivative
        else:
            fiducial = self.fiducial
            derivative = self.derivative
        return fiducial, derivative

    def get_summaries(self, w, key=None, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        Selects either the fitting or validation sets and passes them through
        the network to get the network outputs. For the numerical derivatives,
        the array is first flattened along the batch axis before being passed
        through the model.

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=False
            Whether to get summaries of the validation set

        Returns
        -------
        float(n_s, n_summaries):
            The set of all network outputs used to calculate the covariance
        float(n_d, 2, n_params, n_summaries):
            The outputs of the network of simulations made at perturbed
            parameter values to construct the derivative of the network outputs
            with respect to the model parameters numerically
        """
        d, d_mp = self._collect_input(key, validate=validate)

        x = self.model(w, d)
        x_mp = np.reshape(
            self.model(
                w, d_mp.reshape(
                    (self.n_d * 2 * self.n_params,) + self.input_shape)),
            (self.n_d, 2, self.n_params, self.n_summaries))
        return x, x_mp

    def _construct_derivatives(self, x_mp):
        """Builds derivatives of the network outputs wrt model parameters

        The network outputs from the simulations generated with model parameter
        values above and below the fiducial are subtracted from each other and
        divided by the perturbation size in each model parameter value. The
        axes are swapped such that the derivatives with respect to parameters
        are in the last axis.

        .. math::
            \\frac{\\partial{\\bf x}^i}{\\partial\\theta_\\alpha} =
            \\frac{{\\bf x}^i_{\\alpha^+}-{\\bf x}^i_{\\alpha^+}}{
            \\delta\\theta_\\alpha}

        Parameters
        ----------
        derivatives : float(n_d, 2, n_params, n_summaries)
            The outputs of the network of simulations made at perturbed
            parameter values to construct the derivative of the network outputs
            with respect to the model parameters numerically

        Returns
        -------
        float(n_d, n_summaries, n_params):
            The numerical derivatives of the network ouputs with respect to the
            model parameters
        """
        return np.swapaxes(x_mp[:, 1] - x_mp[:, 0], 1, 2) / self.δθ



from typing import Callable, Union, Sequence
from jax._src.api import _check_callable, _check_arg, vmap, \
    _jvp, _check_input_dtype_jacfwd, _check_output_dtype_jacfwd, \
    _vjp, _check_input_dtype_jacrev, _check_output_dtype_jacrev, \
    _std_basis, _jacfwd_unravel, _jacrev_unravel
from jax._src.api_util import argnums_partial, _ensure_index
from jax.tree_util import tree_map, tree_structure, tree_transpose
from jax._src.util import partial, wraps
from jax._src.traceback_util import api_boundary
import numpy as np
import jax.linear_util as lu
import jax.core as core
import jax.dtypes as dtypes

def jacfwd(fun: Callable, argnums: Union[int, Sequence[int]] = 0,
           holomorphic: bool = False) -> Callable:
  """Creates a function returning Jacobian of ``fun`` evaluated column-by-column
     using forward-mode AD.

  Args:
    fun: Function whose Jacobian is to be computed.
    argnums: Optional, integer or sequence of integers. Specifies which
      positional argument(s) to differentiate with respect to (default ``0``).
    holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
      holomorphic. Default False.

  Returns:
    A function with the same arguments as ``fun``, that evaluates the Jacobian of
    ``fun`` using forward-mode automatic differentiation.

  For example:

  >>> import jax
  >>> import jax.numpy as jnp
  >>>
  >>> def f(x):
  ...   return jnp.asarray(
  ...     [x[0], 5*x[2], 4*x[1]**2 - 2*x[2], x[2] * jnp.sin(x[0])])
  ...
  >>> print(jax.jacfwd(f)(jnp.array([1., 2., 3.])))
  [[ 1.       0.       0.     ]
   [ 0.       0.       5.     ]
   [ 0.      16.      -2.     ]
   [ 1.6209   0.       0.84147]]
  """
  value_and_jacfwd_f = value_and_jacfwd(fun, argnums, holomorphic=holomorphic)

  docstr = ("Foward-mode Jacobian of {fun} with respect to positional "
            "argument(s) {argnums}.")

  @wraps(fun, docstr=docstr, argnums=argnums)
  @api_boundary
  def jacfwd_f(*args, **kwargs):
    _, jf = value_and_jacfwd_f(*args, **kwargs)
    return jf

  return jacfwd_f

def value_and_jacfwd(fun: Callable, argnums: Union[int, Sequence[int]] = 0,
                     holomorphic: bool = False) -> Callable:
  """Jacobian of ``fun`` evaluated column-by-column using forward-mode AD.

  Args:
    fun: Function whose Jacobian is to be computed.
    argnums: Optional, integer or sequence of integers. Specifies which
      positional argument(s) to differentiate with respect to (default ``0``).
    holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
      holomorphic. Default False.

  Returns:
    The value of the function to be automatically differentiated.
    A function with the same arguments as ``fun``, that evaluates the Jacobian of
    ``fun`` using forward-mode automatic differentiation.

  For example:

  >>> import jax
  >>> import jax.numpy as jnpNon-hashable static arguments are not supported, as this can lead to unexpected cache-misses.
  >>> def f(x):
  ...   return jnp.asarray(
  ...     [x[0], 5*x[2], 4*x[1]**2 - 2*x[2], x[2] * jnp.sin(x[0])])
  ...
  >>> print(jax.value_and_jacfwd(f)(jnp.array([1., 2., 3.]))[0])
  [ 1.        15.        10.         2.5244129]
  >>> print(jax.value_and_jacfwd(f)(jnp.array([1., 2., 3.]))[1])
  [[ 1.          0.          0.        ]
   [ 0.          0.          5.        ]
   [ 0.         16.         -2.        ]
   [ 1.6209068   0.          0.84147096]]
  """
  docstr = ("Value and forward-mode Jacobian of {fun} with respect to "
            "positional argument(s) {argnums}. Takes the same arguments as "
            "{fun} but returns a two-element tuple where the first element "
            "is the value of {fun} and the second element is the Jacobian.")

  _check_callable(fun)
  argnums = _ensure_index(argnums)

  @wraps(fun, docstr=docstr, argnums=argnums)
  @api_boundary
  def value_and_jacfwd_f(*args, **kwargs):
    f = lu.wrap_init(fun, kwargs)
    f_partial, dyn_args = argnums_partial(f, argnums, args, require_static_args_hashable=False)
    tree_map(partial(_check_input_dtype_jacfwd, holomorphic), dyn_args)
    pushfwd = partial(_jvp, f_partial, dyn_args)
    y, jac = vmap(pushfwd, out_axes=(None, -1))(_std_basis(dyn_args))
    tree_map(partial(_check_output_dtype_jacfwd, holomorphic), y)
    example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
    return y, tree_map(partial(_jacfwd_unravel, example_args), y, jac)

  return value_and_jacfwd_f

def _check_input_dtype_jacfwd(holomorphic, x):
  _check_arg(x)
  aval = core.get_aval(x)
  if holomorphic:
    if not (dtypes.issubdtype(aval.dtype, np.complexfloating) and
            not dtypes.issubdtype(aval.dtype, np.floating)):
      raise TypeError("jacfwd with holomorphic=True requires inputs with complex dtype, "
                      f"but got {aval.dtype.name}.")
  elif not dtypes.issubdtype(aval.dtype, np.floating):
    raise TypeError("jacfwd requires real-valued inputs (input dtype that is "
                    f"a sub-dtype of np.floating), but got {aval.dtype.name}. "
                    "For holomorphic differentiation, pass holomorphic=True. "
                    "For differentiation of non-holomorphic functions involving complex "
                    "inputs or integer inputs, use jax.jvp directly.")

def _check_output_dtype_jacfwd(holomorphic, x):
  aval = core.get_aval(x)
  if holomorphic:
    if not dtypes.issubdtype(aval.dtype, np.complexfloating):
      raise TypeError("jacfwd with holomorphic=True requires outputs with complex dtype, "
                      f"but got {aval.dtype.name}.")


def jacrev(fun: Callable, argnums: Union[int, Sequence[int]] = 0,
           has_aux: bool = False, holomorphic: bool = False,
           allow_int: bool = False) -> Callable:
  """Creates a function which evaluates the Jacobian of ``fun`` evaluated
     row-by-row using reverse-mode AD.

  Args:
    fun: Function whose Jacobian is to be computed.
    argnums: Optional, integer or sequence of integers. Specifies which
      positional argument(s) to differentiate with respect to (default ``0``).
    has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where the
      first element is considered the output of the mathematical function to be
      differentiated and the second element is auxiliary data. Default False.
    holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
      holomorphic. Default False.
    allow_int: Optional, bool. Whether to allow differentiating with
      respect to integer valued inputs. The gradient of an integer input will
      have a trivial vector-space dtype (float0). Default False.

  Returns:
    A function with the same arguments as ``fun``, that evaluates the Jacobian of
    ``fun`` using reverse-mode automatic differentiation. If ``has_aux`` is True
    then a pair of (Jacobian, auxiliary_data) is returned.

  For example:

  >>> import jax
  >>> import jax.numpy as jnp
  >>>
  >>> def f(x):
  ...   return jnp.asarray(
  ...     [x[0], 5*x[2], 4*x[1]**2 - 2*x[2], x[2] * jnp.sin(x[0])])
  ...
  >>> print(jax.jacrev(f)(jnp.array([1., 2., 3.])))
  [[ 1.       0.       0.     ]
   [ 0.       0.       5.     ]
   [ 0.      16.      -2.     ]
   [ 1.6209   0.       0.84147]]
  """
  value_and_jacrev_f = value_and_jacrev(fun, argnums, has_aux=has_aux,
                                        holomorphic=holomorphic,
                                        allow_int=allow_int)

  docstr = ("Reverse-mode Jacobian of {fun} with respect to positional "
          "argument(s) {argnums}.")

  @wraps(fun, docstr=docstr, argnums=argnums)
  @api_boundary
  def jacrev_f(*args, **kwargs):
    _, jr = value_and_jacrev_f(*args, **kwargs)
    return jr

  @wraps(fun, docstr=docstr, argnums=argnums)
  @api_boundary
  def jacrev_f_aux(*args, **kwargs):
    (_, aux), jr = value_and_jacrev_f(*args, **kwargs)
    return jr, aux

  return jacrev_f_aux if has_aux else jacrev_f

def value_and_jacrev(fun: Callable, argnums: Union[int, Sequence[int]] = 0,
                     has_aux: bool = False, holomorphic: bool = False,
                     allow_int: bool = False) -> Callable:
  """Jacobian of ``fun`` evaluated row-by-row using reverse-mode AD.

  Args:
    fun: Function whose Jacobian is to be computed.
    argnums: Optional, integer or sequence of integers. Specifies which
      positional argument(s) to differentiate with respect to (default ``0``).
    has_aux: Optional, bool. Indicates whether ``fun`` returns a pair where the
      first element is considered the output of the mathematical function to be
      differentiated and the second element is auxiliary data. Default False.
    holomorphic: Optional, bool. Indicates whether ``fun`` is promised to be
      holomorphic. Default False.
    allow_int: Optional, bool. Whether to allow differentiating with
      respect to integer valued inputs. The gradient of an integer input will
      have a trivial vector-space dtype (float0). Default False.

  Returns:
    The value of the function to be automatically differentiated.
    A function with the same arguments as ``fun``, that evaluates the Jacobian of
    ``fun`` using reverse-mode automatic differentiation.

  For example:

  >>> import jax
  >>> import jax.numpy as jnp
  >>>
  >>> def f(x):
  ...   return jnp.asarray(
  ...     [x[0], 5*x[2], 4*x[1]**2 - 2*x[2], x[2] * jnp.sin(x[0])])
  ...
  >>> print(jax.value_and_jacrev(f)(jnp.array([1., 2., 3.]))[0])
  [ 1.        15.        10.         2.5244129]
  >>> print(jax.value_and_jacrev(f)(jnp.array([1., 2., 3.]))[1])
  [[ 1.          0.          0.        ]
   [ 0.          0.          5.        ]
   [ 0.         16.         -2.        ]
   [ 1.6209068   0.          0.84147096]]
  """

  docstr = ("Value and reverse-mode Jacobian of {fun} with respect to "
            "positional argument(s) {argnums}. Takes the same arguments as "
            "{fun} but returns a two-element tuple where the first element "
            "is the value of {fun} and the second element is the Jacobian.")

  _check_callable(fun)

  @wraps(fun, docstr=docstr, argnums=argnums)
  @api_boundary
  def value_and_jacrev_f(*args, **kwargs):
    f = lu.wrap_init(fun, kwargs)
    f_partial, dyn_args = argnums_partial(f, argnums, args, require_static_args_hashable=False)
    tree_map(partial(_check_input_dtype_jacrev, holomorphic, allow_int), dyn_args)
    if not has_aux:
      y, pullback = _vjp(f_partial, *dyn_args)
    else:
      y, pullback, aux = _vjp(f_partial, *dyn_args, has_aux=True)
    tree_map(partial(_check_output_dtype_jacrev, holomorphic), y)
    jac = vmap(pullback)(_std_basis(y))
    jac = jac[0] if isinstance(argnums, int) else jac
    example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
    jac_tree = tree_map(partial(_jacrev_unravel, y), example_args, jac)
    if not has_aux:
      return y, tree_transpose(tree_structure(example_args), tree_structure(y), jac_tree)
    else:
      return (y, aux), tree_transpose(tree_structure(example_args), tree_structure(y), jac_tree)
    return

  return value_and_jacrev_f



import pytest
import jax
import tensorflow as tf
from imnn import IMNN, SimulatorIMNN, AggregatedSimulatorIMNN, \
    GradientIMNN, AggregatedGradientIMNN, DatasetGradientIMNN, \
    NumericalGradientIMNN, AggregatedNumericalGradientIMNN, \
    DatasetNumericalGradientIMNN
from test.defaults import defaultTests
from test.simulator_defaults import simulatorTests
from test.gradient_defaults import gradientTests
from test.numerical_gradient_defaults import numericalGradientTests

defaults = defaultTests()
simulator_defaults = simulatorTests()
gradient_defaults = gradientTests()
numerical_gradient_defaults = numericalGradientTests()

host = jax.devices("cpu")[0]
devices = jax.devices()
n_devices = len(devices)
n_per_device = defaults.n_d


@pytest.mark.parametrize(
    "model",
    [defaults.model,
     tf.keras.models.Sequential([
         tf.keras.layers.Dense(128),
         tf.keras.layers.LeakyReLU(0.01),
         tf.keras.layers.Dense(128),
         tf.keras.layers.LeakyReLU(0.01),
         tf.keras.layers.Dense(128),
         tf.keras.layers.LeakyReLU(0.01),
         tf.keras.layers.Dense(defaults.n_d)])
     ])
@pytest.mark.parametrize(
    "optimiser",
    [defaults.optimiser,
     tf.keras.optimizers.Adam()
     ])
@pytest.mark.parametrize("simulator", [None, simulator_defaults.simulator])
@pytest.mark.parametrize(
    "fiducial",
    [None, gradient_defaults.fiducial, numerical_gradient_defaults.fiducial,
     [tf.data.Dataset.from_tensor_slices(
         fid).repeat().as_numpy_iterator()
      for fid in numerical_gradient_defaults.fiducial.reshape(
          (n_devices,
           defaults.n_s // (n_devices * n_per_device),
           n_per_device)
          + defaults.input_shape)]])
@pytest.mark.parametrize(
    "derivative",
    [None, gradient_defaults.derivative,
     numerical_gradient_defaults.derivative,
     [tf.data.Dataset.from_tensor_slices(
         der).repeat().as_numpy_iterator()
      for der in numerical_gradient_defaults.derivative.reshape(
          (n_devices,
           2 * defaults.n_params * defaults.n_d //
              (n_devices * n_per_device),
           n_per_device)
          + defaults.input_shape)]])
@pytest.mark.parametrize(
    "main",
    [None,
     [tf.data.Dataset.zip((
         tf.data.Dataset.from_tensor_slices(fid),
         tf.data.Dataset.from_tensor_slices(der))).repeat().as_numpy_iterator()
      for fid, der in zip(
          gradient_defaults.fiducial[:defaults.n_d].reshape(
              (n_devices,
               defaults.n_d // (n_devices * n_per_device),
               n_per_device)
              + defaults.input_shape),
          gradient_defaults.derivative.reshape(
              (n_devices,
               defaults.n_d // (n_devices * n_per_device),
               n_per_device)
              + defaults.input_shape
              + (defaults.n_params,)))]])
@pytest.mark.parametrize(
    "remaining",
    [None,
     [tf.data.Dataset.from_tensor_slices(
         fid).repeat().as_numpy_iterator()
      for fid in gradient_defaults.fiducial[defaults.n_d:].reshape(
          (n_devices,
           (defaults.n_s - defaults.n_d) // (n_devices * n_per_device),
           n_per_device)
          + defaults.input_shape)]])
@pytest.mark.parametrize(
    "δθ",
    [None, numerical_gradient_defaults.δθ])
@pytest.mark.parametrize(
    "validation_fiducial",
    [None,
     gradient_defaults.validation_fiducial,
     numerical_gradient_defaults.validation_fiducial,
     [tf.data.Dataset.from_tensor_slices(
         fid).repeat().as_numpy_iterator()
      for fid in numerical_gradient_defaults.validation_fiducial.reshape(
          (n_devices,
           defaults.n_s // (n_devices * n_per_device),
           n_per_device)
          + defaults.input_shape)]])
@pytest.mark.parametrize(
    "validation_derivative",
    [None,
     gradient_defaults.validation_derivative,
     numerical_gradient_defaults.validation_derivative,
     [tf.data.Dataset.from_tensor_slices(
         der).repeat().as_numpy_iterator()
      for der in numerical_gradient_defaults.validation_derivative.reshape(
          (n_devices,
           2 * defaults.n_params * defaults.n_d //
              (n_devices * n_per_device),
           n_per_device)
          + defaults.input_shape)]])
@pytest.mark.parametrize(
    "validation_main",
    [None,
     [tf.data.Dataset.zip((
         tf.data.Dataset.from_tensor_slices(fid),
         tf.data.Dataset.from_tensor_slices(der))).repeat().as_numpy_iterator()
      for fid, der in zip(
          gradient_defaults.validation_fiducial[:defaults.n_d].reshape(
              (n_devices,
               defaults.n_d // (n_devices * n_per_device),
               n_per_device)
              + defaults.input_shape),
          gradient_defaults.validation_derivative.reshape(
              (n_devices,
               defaults.n_d // (n_devices * n_per_device),
               n_per_device)
              + defaults.input_shape
              + (defaults.n_params,)))]])
@pytest.mark.parametrize(
    "validation_remaining",
    [None,
     [tf.data.Dataset.from_tensor_slices(
         fid).repeat().as_numpy_iterator()
      for fid in gradient_defaults.validation_fiducial[defaults.n_d:].reshape(
          (n_devices,
           (defaults.n_s - defaults.n_d) // (n_devices * n_per_device),
           n_per_device)
          + defaults.input_shape)]])
@pytest.mark.parametrize("host", [host])
@pytest.mark.parametrize("devices", [devices])
@pytest.mark.parametrize("n_per_device", [n_per_device])
@pytest.mark.parametrize("cache", [None, False])
@pytest.mark.parametrize("prefetch", [None, tf.data.AUTOTUNE])
def test_IMNN(
        model, optimiser, simulator, fiducial, derivative, main, remaining, δθ,
        validation_fiducial, validation_derivative, validation_main,
        validation_remaining, host, devices, n_per_device, cache, prefetch):
    def call_IMNN():
        return IMNN(
            n_s=defaults.n_s,
            n_d=defaults.n_d,
            n_params=defaults.n_params,
            n_summaries=defaults.n_summaries,
            input_shape=defaults.input_shape,
            θ_fid=defaults.θ_fid,
            model=model,
            optimiser=optimiser,
            key_or_state=defaults.model_key,
            simulator=simulator,
            fiducial=fiducial,
            derivative=derivative,
            main=main,
            remaining=remaining,
            δθ=δθ,
            validation_fiducial=validation_fiducial,
            validation_derivative=validation_derivative,
            validation_main=validation_main,
            validation_remaining=validation_remaining,
            host=host,
            devices=devices,
            n_per_device=n_per_device,
            cache=cache,
            prefetch=prefetch)

    numerical_shape = (defaults.n_d, 2, defaults.n_params) \
        + defaults.input_shape
    gradient_shape = (defaults.n_d,) + defaults.input_shape \
        + (defaults.n_params,)

    if (isinstance(model, tf.keras.models.Model)
            or isinstance(optimiser, tf.keras.optimizers.Optimizer)):
        with pytest.raises(ValueError):
            call_IMNN()
        return

    if ((simulator is not None)
            and (fiducial is None)
            and (derivative is None)
            and (main is None)
            and (remaining is None)
            and (δθ is None)
            and (validation_fiducial is None)
            and (validation_derivative is None)
            and (validation_main is None)
            and (validation_remaining is None)
            and (cache is None)
            and (prefetch is None)):
        if (host is None) and (devices is None) and (n_per_device is None):
            assert isinstance(call_IMNN(), SimulatorIMNN)
            return
        if ((host is not None)
                and (devices is not None)
                and (n_per_device is not None)):
            assert isinstance(call_IMNN(), AggregatedSimulatorIMNN)
            return

    if ((fiducial is not None)
            and (derivative is not None)
            and (simulator is None)
            and (((host is None)
                  and (devices is None)
                  and (n_per_device is None)
                  and (cache is None)
                  and (prefetch is None))
                 or ((host is not None)
                     and (devices is not None)
                     and (n_per_device is not None)))):
        if (any([isinstance(fiducial, list),
                 isinstance(derivative, list)])
                and (not all([isinstance(fiducial, list),
                              isinstance(derivative, list)]))):
            with pytest.raises(TypeError):
                call_IMNN()
            return
        if ((δθ is not None)
                and (((validation_fiducial is None)
                      and (validation_derivative is not None))
                     or ((validation_fiducial is not None)
                         and (validation_derivative is None)))):
            with pytest.raises(ValueError):
                call_IMNN()
            return
        if all([isinstance(fiducial, list), isinstance(derivative, list)]):
            if ((δθ is None)
                    or ((host is None)
                        and (devices is None)
                        and (n_per_device is None))):
                with pytest.raises(TypeError):
                    call_IMNN()
                return
        if (((validation_fiducial is not None)
                and (validation_derivative is not None))
                and any([isinstance(fiducial, list),
                         isinstance(derivative, list),
                         isinstance(validation_fiducial, list),
                         isinstance(validation_derivative, list)])
                and (not all([isinstance(fiducial, list),
                              isinstance(derivative, list),
                              isinstance(validation_fiducial, list),
                              isinstance(validation_derivative, list)]))):
            if ((main is not None)
                    or (remaining is not None)
                    or (validation_main is not None)
                    or (validation_remaining is not None)):
                with pytest.raises(ValueError):
                    call_IMNN()
                return
            else:
                with pytest.raises(TypeError):
                    call_IMNN()
                return

    if ((fiducial is not None)
            and (derivative is not None)
            and (main is None)
            and (remaining is None)
            and (validation_main is None)
            and (validation_remaining is None)
            and (not isinstance(model, tf.keras.models.Model))
            and (not isinstance(optimiser, tf.keras.optimizers.Optimizer))
            and (simulator is None)):
        if ((validation_fiducial is not None)
                and (validation_derivative is not None)):
            if δθ is not None:
                if (isinstance(fiducial, list)
                        and isinstance(derivative, list)
                        and isinstance(validation_fiducial, list)
                        and isinstance(validation_derivative, list)
                        and (cache is None)
                        and (prefetch is None)
                        and (host is not None)
                        and (devices is not None)
                        and (n_per_device is not None)):
                    assert isinstance(
                        call_IMNN(), DatasetNumericalGradientIMNN)
                    return
                if not any([isinstance(fiducial, list),
                            isinstance(derivative, list),
                            isinstance(validation_fiducial, list),
                            isinstance(validation_derivative, list)]):
                    if ((validation_derivative.shape == numerical_shape)
                            and (derivative.shape == numerical_shape)):
                        if ((host is None)
                                and (devices is None)
                                and (n_per_device is None)
                                and (cache is None)
                                and (prefetch is None)):
                            assert isinstance(
                                call_IMNN(), NumericalGradientIMNN)
                            return
                        if ((host is not None)
                                and (devices is not None)
                                and (n_per_device is not None)):
                            assert isinstance(
                                call_IMNN(), AggregatedNumericalGradientIMNN)
                            return
            else:
                if not any([isinstance(fiducial, list),
                            isinstance(derivative, list),
                            isinstance(validation_fiducial, list),
                            isinstance(validation_derivative, list)]):
                    if ((validation_derivative.shape == gradient_shape)
                            and (derivative.shape == gradient_shape)):
                        if ((host is None)
                                and (devices is None)
                                and (n_per_device is None)
                                and (cache is None)
                                and (prefetch is None)):
                            assert isinstance(call_IMNN(), GradientIMNN)
                            return
                        if ((host is not None)
                                and (devices is not None)
                                and (n_per_device is not None)):
                            assert isinstance(
                                call_IMNN(), AggregatedGradientIMNN)
                            return
        if ((validation_fiducial is None) and (validation_derivative is None)):
            if δθ is not None:
                if (isinstance(fiducial, list)
                        and isinstance(derivative, list)
                        and (cache is None)
                        and (prefetch is None)
                        and (host is not None)
                        and (devices is not None)
                        and (n_per_device is not None)):
                    assert isinstance(
                        call_IMNN(), DatasetNumericalGradientIMNN)
                    return
                if not any([isinstance(fiducial, list),
                            isinstance(derivative, list),
                            isinstance(validation_fiducial, list),
                            isinstance(validation_derivative, list)]):
                    if derivative.shape == numerical_shape:
                        if ((host is None)
                                and (devices is None)
                                and (n_per_device is None)
                                and (cache is None)
                                and (prefetch is None)):
                            assert isinstance(
                                call_IMNN(), NumericalGradientIMNN)
                            return
                        if ((host is not None)
                                and (devices is not None)
                                and (n_per_device is not None)):
                            assert isinstance(
                                call_IMNN(), AggregatedNumericalGradientIMNN)
                            return
            else:
                if not any([isinstance(fiducial, list),
                            isinstance(derivative, list),
                            isinstance(validation_fiducial, list),
                            isinstance(validation_derivative, list)]):
                    if derivative.shape == gradient_shape:
                        if ((host is None)
                                and (devices is None)
                                and (n_per_device is None)
                                and (cache is None)
                                and (prefetch is None)):
                            assert isinstance(
                                call_IMNN(), GradientIMNN)
                            return
                        if ((host is not None)
                                and (devices is not None)
                                and (n_per_device is not None)):
                            assert isinstance(
                                call_IMNN(), AggregatedGradientIMNN)
                            return
    if ((main is not None)
            and (remaining is not None)
            and (fiducial is None)
            and (derivative is None)
            and (validation_fiducial is None)
            and (validation_derivative is None)
            and (simulator is None)
            and (prefetch is None)
            and (cache is None)
            and (δθ is None)
            and (host is not None)
            and (devices is not None)
            and (n_per_device is not None)):
        if (((validation_main is not None)
                and (validation_remaining is not None))
                or ((validation_main is None)
                    and (validation_remaining is None))):
            assert isinstance(call_IMNN(), DatasetGradientIMNN)
            return

    with pytest.raises(ValueError):
        call_IMNN()
